# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""SPM wrappers for preprocessing data
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import range

import os
from copy import deepcopy

# Third-party imports
import numpy as np

# Local imports
from ...utils.filemanip import (fname_presuffix, ensure_list,
                                simplify_list, split_filename)
from ..base import (OutputMultiPath, TraitedSpec, isdefined,
                    traits, InputMultiPath, InputMultiObject, File, Str)
from .base import (SPMCommand, scans_for_fname, func_is_3d,
                   scans_for_fnames, SPMCommandInputSpec, ImageFileSPM)

__docformat__ = 'restructuredtext'


class FieldMapInputSpec(SPMCommandInputSpec):
    jobtype = traits.Enum('calculatevdm', 'applyvdm', usedefault=True,
        desc='one of: calculatevdm, applyvdm')
    phase_file = File(mandatory=True, exists=True, copyfile=False,
        field='subj.data.presubphasemag.phase',
        desc='presubstracted phase file')
    magnitude_file = File(mandatory=True, exists=True, copyfile=False,
        field='subj.data.presubphasemag.magnitude',
        desc='presubstracted magnitude file')
    echo_times = traits.Tuple(traits.Float, traits.Float, mandatory=True,
        field='subj.defaults.defaultsval.et',
        desc='short and long echo times')
    maskbrain = traits.Bool(True, usedefault=True,
        field='subj.defaults.defaultsval.maskbrain',
        desc='masking or no masking of the brain')
    blip_direction = traits.Enum(1, -1, mandatory=True,
        field='subj.defaults.defaultsval.blipdir',
        desc='polarity of the phase-encode blips')
    total_readout_time = traits.Float(mandatory=True,
        field='subj.defaults.defaultsval.tert',
        desc='total EPI readout time')
    epifm = traits.Bool(False, usedefault=True,
        field='subj.defaults.defaultsval.epifm',
        desc='epi-based field map');
    jacobian_modulation = traits.Bool(False, usedefault=True,
        field='subj.defaults.defaultsval.ajm',
        desc='jacobian modulation');
    # Unwarping defaults parameters
    method = traits.Enum('Mark3D', 'Mark2D', 'Huttonish', usedefault=True,
        desc='One of: Mark3D, Mark2D, Huttonish',
        field='subj.defaults.defaultsval.uflags.method');
    unwarp_fwhm = traits.Range(low=0, value=10, usedefault=True,
        field='subj.defaults.defaultsval.uflags.fwhm',
        desc='gaussian smoothing kernel width');
    pad = traits.Range(low=0, value=0, usedefault=True,
        field='subj.defaults.defaultsval.uflags.pad',
        desc='padding kernel width');
    ws = traits.Bool(True, usedefault=True,
        field='subj.defaults.defaultsval.uflags.ws',
        desc='weighted smoothing');
    # Brain mask defaults parameters
    template = File(copyfile=False, exists=True,
        field='subj.defaults.defaultsval.mflags.template',
        desc='template image for brain masking');
    mask_fwhm = traits.Range(low=0, value=5, usedefault=True,
        field='subj.defaults.defaultsval.mflags.fwhm',
        desc='gaussian smoothing kernel width');
    nerode = traits.Range(low=0, value=2, usedefault=True,
        field='subj.defaults.defaultsval.mflags.nerode',
        desc='number of erosions');
    ndilate = traits.Range(low=0, value=4, usedefault=True,
        field='subj.defaults.defaultsval.mflags.ndilate',
        desc='number of erosions');
    thresh = traits.Float(0.5, usedefault=True,
        field='subj.defaults.defaultsval.mflags.thresh',
        desc='threshold used to create brain mask from segmented data');
    reg = traits.Float(0.02, usedefault=True,
        field='subj.defaults.defaultsval.mflags.reg',
        desc='regularization value used in the segmentation');
    # EPI unwarping for quality check
    epi_file = File(copyfile=False, exists=True, mandatory=True,
        field='subj.session.epi',
        desc='EPI to unwarp');
    matchvdm = traits.Bool(True, usedefault=True,
        field='subj.matchvdm',
        desc='match VDM to EPI');
    sessname = Str('_run-', usedefault=True,
        field='subj.sessname',
        desc='VDM filename extension');
    writeunwarped = traits.Bool(False, usedefault=True,
        field='subj.writeunwarped',
        desc='write unwarped EPI');
    anat_file = File(copyfile=False, exists=True,
        field='subj.anat',
        desc='anatomical image for comparison');
    matchanat = traits.Bool(True, usedefault=True,
        field='subj.matchanat',
        desc='match anatomical image to EPI');


class FieldMapOutputSpec(TraitedSpec):
    vdm = File(exists=True, desc='voxel difference map')


class FieldMap(SPMCommand):
    """Use the fieldmap toolbox from spm to calculate the voxel displacement map (VDM).

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=173

    To do
    -----
    Deal with real/imag magnitude images and with the two phase files case.

    Examples
    --------
    >>> from nipype.interfaces.spm import FieldMap
    >>> fm = FieldMap()
    >>> fm.inputs.phase_file = 'phase.nii'
    >>> fm.inputs.magnitude_file = 'magnitude.nii'
    >>> fm.inputs.echo_times = (5.19, 7.65)
    >>> fm.inputs.blip_direction = 1
    >>> fm.inputs.total_readout_time = 15.6
    >>> fm.inputs.epi_file = 'epi.nii'
    >>> fm.run() # doctest: +SKIP

    """

    input_spec = FieldMapInputSpec
    output_spec = FieldMapOutputSpec
    _jobtype = 'tools'
    _jobname = 'fieldmap'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ['phase_file', 'magnitude_file', 'anat_file', 'epi_file']:
            return scans_for_fname(ensure_list(val))

        return super(FieldMap, self)._format_arg(opt, spec, val)

    def _parse_inputs(self):
        """validate spm fieldmap options if set to None ignore
        """
        einputs = super(FieldMap, self)._parse_inputs()
        return [{self.inputs.jobtype: einputs[0]}]

    def _list_outputs(self):
        outputs = self._outputs().get()
        jobtype = self.inputs.jobtype
        if jobtype == "calculatevdm":
            outputs['vdm'] = fname_presuffix(self.inputs.phase_file, prefix='vdm5_sc')

        return outputs


class SliceTimingInputSpec(SPMCommandInputSpec):
    in_files = InputMultiPath(
        traits.Either(
            traits.List(ImageFileSPM(exists=True)), ImageFileSPM(exists=True)),
        field='scans',
        desc='list of filenames to apply slice timing',
        mandatory=True,
        copyfile=False)
    num_slices = traits.Int(
        field='nslices', desc='number of slices in a volume', mandatory=True)
    time_repetition = traits.Float(
        field='tr',
        desc=('time between volume acquisitions'
              '(start to start time)'),
        mandatory=True)
    time_acquisition = traits.Float(
        field='ta',
        desc=('time of volume acquisition. usually'
              'calculated as TR-(TR/num_slices)'),
        mandatory=True)
    slice_order = traits.List(
        traits.Float(),
        field='so',
        desc=('1-based order or onset (in ms) in which '
              'slices are acquired'),
        mandatory=True)
    ref_slice = traits.Int(
        field='refslice',
        desc='1-based Number of the reference slice or '
        'reference time point if slice_order is in '
        'onsets (ms)',
        mandatory=True)
    out_prefix = traits.String(
        'a', field='prefix', usedefault=True, desc='slicetimed output prefix')


class SliceTimingOutputSpec(TraitedSpec):
    timecorrected_files = OutputMultiPath(
        traits.Either(traits.List(File(exists=True)), File(exists=True)),
        desc='slice time corrected files')


class SliceTiming(SPMCommand):
    """Use spm to perform slice timing correction.

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=19

    Examples
    --------

    >>> from nipype.interfaces.spm import SliceTiming
    >>> st = SliceTiming()
    >>> st.inputs.in_files = 'functional.nii'
    >>> st.inputs.num_slices = 32
    >>> st.inputs.time_repetition = 6.0
    >>> st.inputs.time_acquisition = 6. - 6./32.
    >>> st.inputs.slice_order = list(range(32,0,-1))
    >>> st.inputs.ref_slice = 1
    >>> st.run() # doctest: +SKIP

    """

    input_spec = SliceTimingInputSpec
    output_spec = SliceTimingOutputSpec

    _jobtype = 'temporal'
    _jobname = 'st'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt == 'in_files':
            return scans_for_fnames(
                ensure_list(val), keep4d=False, separate_sessions=True)
        return super(SliceTiming, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['timecorrected_files'] = []

        filelist = ensure_list(self.inputs.in_files)
        for f in filelist:
            if isinstance(f, list):
                run = [
                    fname_presuffix(in_f, prefix=self.inputs.out_prefix)
                    for in_f in f
                ]
            else:
                run = fname_presuffix(f, prefix=self.inputs.out_prefix)
            outputs['timecorrected_files'].append(run)
        return outputs


class RealignInputSpec(SPMCommandInputSpec):
    in_files = InputMultiPath(
        traits.Either(ImageFileSPM(exists=True),
                      traits.List(ImageFileSPM(exists=True))),
        field='data',
        mandatory=True,
        copyfile=True,
        desc='list of filenames to realign')
    jobtype = traits.Enum(
        'estwrite',
        'estimate',
        'write',
        desc='one of: estimate, write, estwrite',
        usedefault=True)
    quality = traits.Range(
        low=0.0,
        high=1.0,
        field='eoptions.quality',
        desc='0.1 = fast, 1.0 = precise')
    fwhm = traits.Range(
        low=0.0, field='eoptions.fwhm', desc='gaussian smoothing kernel width')
    separation = traits.Range(
        low=0.0, field='eoptions.sep', desc='sampling separation in mm')
    register_to_mean = traits.Bool(
        field='eoptions.rtm',
        desc=('Indicate whether realignment is '
              'done to the mean image'))
    weight_img = File(
        exists=True,
        field='eoptions.weight',
        desc='filename of weighting image')
    interp = traits.Range(
        low=0,
        high=7,
        field='eoptions.interp',
        desc='degree of b-spline used for interpolation')
    wrap = traits.List(
        traits.Int(),
        minlen=3,
        maxlen=3,
        field='eoptions.wrap',
        desc='Check if interpolation should wrap in [x,y,z]')
    write_which = traits.ListInt(
        [2, 1],
        field='roptions.which',
        minlen=2,
        maxlen=2,
        usedefault=True,
        desc='determines which images to reslice')
    write_interp = traits.Range(
        low=0,
        high=7,
        field='roptions.interp',
        desc=('degree of b-spline used for '
              'interpolation'))
    write_wrap = traits.List(
        traits.Int(),
        minlen=3,
        maxlen=3,
        field='roptions.wrap',
        desc=('Check if interpolation should wrap in '
              '[x,y,z]'))
    write_mask = traits.Bool(
        field='roptions.mask', desc='True/False mask output image')
    out_prefix = traits.String(
        'r',
        field='roptions.prefix',
        usedefault=True,
        desc='realigned output prefix')


class RealignOutputSpec(TraitedSpec):
    mean_image = File(exists=True, desc='Mean image file from the realignment')
    modified_in_files = OutputMultiPath(
        traits.Either(traits.List(File(exists=True)), File(exists=True)),
        desc=('Copies of all files passed to '
              'in_files. Headers will have '
              'been modified to align all '
              'images with the first, or '
              'optionally to first do that, '
              'extract a mean image, and '
              're-align to that mean image.'))
    realigned_files = OutputMultiPath(
        traits.Either(traits.List(File(exists=True)), File(exists=True)),
        desc=('If jobtype is write or estwrite, '
              'these will be the resliced files.'
              ' Otherwise, they will be copies '
              'of in_files that have had their '
              'headers rewritten.'))
    realignment_parameters = OutputMultiPath(
        File(exists=True),
        desc=('Estimated translation and '
              'rotation parameters'))


class Realign(SPMCommand):
    """Use spm_realign for estimating within modality rigid body alignment

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=25

    Examples
    --------

    >>> import nipype.interfaces.spm as spm
    >>> realign = spm.Realign()
    >>> realign.inputs.in_files = 'functional.nii'
    >>> realign.inputs.register_to_mean = True
    >>> realign.run() # doctest: +SKIP

    """

    input_spec = RealignInputSpec
    output_spec = RealignOutputSpec

    _jobtype = 'spatial'
    _jobname = 'realign'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt == 'in_files':
            if self.inputs.jobtype == "write":
                separate_sessions = False
            else:
                separate_sessions = True
            return scans_for_fnames(
                val, keep4d=False, separate_sessions=separate_sessions)
        return super(Realign, self)._format_arg(opt, spec, val)

    def _parse_inputs(self):
        """validate spm realign options if set to None ignore
        """
        einputs = super(Realign, self)._parse_inputs()
        return [{'%s' % (self.inputs.jobtype): einputs[0]}]

    def _list_outputs(self):
        outputs = self._outputs().get()
        resliced_all = self.inputs.write_which[0] > 0
        resliced_mean = self.inputs.write_which[1] > 0

        if self.inputs.jobtype != "write":
            if isdefined(self.inputs.in_files):
                outputs['realignment_parameters'] = []
            for imgf in self.inputs.in_files:
                if isinstance(imgf, list):
                    tmp_imgf = imgf[0]
                else:
                    tmp_imgf = imgf
                outputs['realignment_parameters'].append(
                    fname_presuffix(
                        tmp_imgf, prefix='rp_', suffix='.txt', use_ext=False))
                if not isinstance(imgf, list) and func_is_3d(imgf):
                    break
        if self.inputs.jobtype == "estimate":
            outputs['realigned_files'] = self.inputs.in_files
        if (self.inputs.jobtype == "estimate"
                or self.inputs.jobtype == "estwrite"):
            outputs['modified_in_files'] = self.inputs.in_files
        if self.inputs.jobtype == "write" or self.inputs.jobtype == "estwrite":
            if isinstance(self.inputs.in_files[0], list):
                first_image = self.inputs.in_files[0][0]
            else:
                first_image = self.inputs.in_files[0]

            if resliced_mean:
                outputs['mean_image'] = fname_presuffix(
                    first_image, prefix='mean')

            if resliced_all:
                outputs['realigned_files'] = []
                for idx, imgf in enumerate(
                        ensure_list(self.inputs.in_files)):
                    realigned_run = []
                    if isinstance(imgf, list):
                        for i, inner_imgf in enumerate(ensure_list(imgf)):
                            newfile = fname_presuffix(
                                inner_imgf, prefix=self.inputs.out_prefix)
                            realigned_run.append(newfile)
                    else:
                        realigned_run = fname_presuffix(
                            imgf, prefix=self.inputs.out_prefix)
                    outputs['realigned_files'].append(realigned_run)
        return outputs


class RealignUnwarpInputSpec(SPMCommandInputSpec):

    in_files = InputMultiObject(
        traits.Either(ImageFileSPM(exists=True),
                      traits.List(ImageFileSPM(exists=True))),
        field='data.scans',
        mandatory=True,
        copyfile=True,
        desc='list of filenames to realign and unwarp')
    phase_map = File(
        field='data.pmscan',
        desc='Voxel displacement map to use in unwarping. Unlike SPM standard '
             'behaviour, the same map will be used for all sessions',
        copyfile=False)
    quality = traits.Range(
        low=0.0,
        high=1.0,
        field='eoptions.quality',
        desc='0.1 = fast, 1.0 = precise')
    fwhm = traits.Range(
        low=0.0,
        field='eoptions.fwhm',
        desc='gaussian smoothing kernel width')
    separation = traits.Range(
        low=0.0,
        field='eoptions.sep',
        desc='sampling separation in mm')
    register_to_mean = traits.Bool(
        field='eoptions.rtm',
        desc='Indicate whether realignment is done to the mean image')
    weight_img = File(
        exists=True,
        field='eoptions.weight',
        desc='filename of weighting image')
    interp = traits.Range(
        low=0,
        high=7,
        field='eoptions.einterp',
        desc='degree of b-spline used for interpolation')
    wrap = traits.List(
        traits.Int(),
        minlen=3,
        maxlen=3,
        field='eoptions.ewrap',
        desc='Check if interpolation should wrap in [x,y,z]')
    est_basis_func = traits.List(
        traits.Int(),
        minlen=2,
        maxlen=2,
        field='uweoptions.basfcn',
        desc='Number of basis functions to use for each dimension')
    est_reg_order = traits.Range(
        low=0,
        high=3,
        field='uweoptions.regorder',
        desc=('This parameter determines how to balance the compromise between likelihood '
              'maximization and smoothness maximization of the estimated field.'))
    est_reg_factor = traits.ListInt(
        [100000],
        field='uweoptions.lambda',
        minlen=1,
        maxlen=1,
        usedefault=True,
        desc='Regularisation factor. Default: 100000 (medium).')
    est_jacobian_deformations = traits.Bool(
        field='uweoptions.jm',
        desc=('Jacobian deformations. In theory a good idea to include them, '
             ' in practice a bad idea. Default: No.'))
    est_first_order_effects = traits.List(
        traits.Int(),
        minlen=1,
        maxlen=6,
        field='uweoptions.fot',
        desc='First order effects should only depend on pitch and roll, i.e. [4 5]')
    est_second_order_effects = traits.List(
        traits.Int(),
        minlen=1,
        maxlen=6,
        field='uweoptions.sot',
        desc='List of second order terms to model second derivatives of.')
    est_unwarp_fwhm = traits.Range(
        low=0.0,
        field='uweoptions.uwfwhm',
        desc='gaussian smoothing kernel width for unwarp')
    est_re_est_mov_par = traits.Bool(
        field='uweoptions.rem',
        desc='Re-estimate movement parameters at each unwarping iteration.')
    est_num_of_iterations = traits.ListInt(
        [5],
        field='uweoptions.noi',
        minlen=1,
        maxlen=1,
        usedefault=True,
        desc='Number of iterations.')
    est_taylor_expansion_point = traits.String(
        'Average',
        field='uweoptions.expround',
        usedefault=True,
        desc='Point in position space to perform Taylor-expansion around.')
    reslice_which = traits.ListInt(
        [2, 1],
        field='uwroptions.uwwhich',
        minlen=2,
        maxlen=2,
        usedefault=True,
        desc='determines which images to reslice')
    reslice_interp = traits.Range(
        low=0,
        high=7,
        field='uwroptions.rinterp',
        desc='degree of b-spline used for interpolation')
    reslice_wrap = traits.List(
        traits.Int(),
        minlen=3,
        maxlen=3,
        field='uwroptions.wrap',
        desc='Check if interpolation should wrap in [x,y,z]')
    reslice_mask = traits.Bool(
        field='uwroptions.mask',
        desc='True/False mask output image')
    out_prefix = traits.String(
        'u',
        field='uwroptions.prefix',
        usedefault=True,
        desc='realigned and unwarped output prefix')


class RealignUnwarpOutputSpec(TraitedSpec):
    mean_image = File(exists=True, desc='Mean image file from the realignment & unwarping')
    modified_in_files = OutputMultiPath(
        traits.Either(traits.List(File(exists=True)), File(exists=True)),
        desc=('Copies of all files passed to '
              'in_files. Headers will have '
              'been modified to align all '
              'images with the first, or '
              'optionally to first do that, '
              'extract a mean image, and '
              're-align to that mean image.'))
    realigned_unwarped_files = OutputMultiPath(
        traits.Either(traits.List(File(exists=True)), File(exists=True)),
        desc='Realigned and unwarped files written to disc.')
    realignment_parameters = OutputMultiPath(
        File(exists=True),
        desc='Estimated translation and rotation parameters')


class RealignUnwarp(SPMCommand):
    """Use spm_uw_estimate for estimating within subject registration and unwarping
    of time series. Function accepts only one single field map. If in_files is a
    list of files they will be treated as separate sessions but associated to the
    same fieldmap.

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=31

    Examples
    --------

    >>> import nipype.interfaces.spm as spm
    >>> realignUnwarp = spm.RealignUnwarp()
    >>> realignUnwarp.inputs.in_files = ['functional.nii', 'functional2.nii']
    >>> realignUnwarp.inputs.phase_map = 'voxeldisplacemap.vdm'
    >>> realignUnwarp.inputs.register_to_mean = True
    >>> realignUnwarp.run() # doctest: +SKIP

    """

    input_spec = RealignUnwarpInputSpec
    output_spec = RealignUnwarpOutputSpec

    _jobtype = 'spatial'
    _jobname = 'realignunwarp'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt == 'in_files':
            return scans_for_fnames(ensure_list(val),
                                    keep4d=False,
                                    separate_sessions=True)
        return super(RealignUnwarp, self)._format_arg(opt, spec, val)


    def _parse_inputs(self, skip=()):

        spmdict = super(RealignUnwarp, self)._parse_inputs(skip=())[0]

        if isdefined(self.inputs.phase_map):
            pmscan = spmdict['data']['pmscan']
        else:
            pmscan = ''

        if isdefined(self.inputs.in_files):
            if isinstance(self.inputs.in_files, list):
                data = [dict(scans = sess, pmscan = pmscan)
                                          for sess in spmdict['data']['scans']]
            else:
                data = [dict(scans = spmdict['data']['scans'], pmscan = pmscan)]

        spmdict['data'] = data

        return [spmdict]


    def _list_outputs(self):
        outputs = self._outputs().get()
        resliced_all = self.inputs.reslice_which[0] > 0
        resliced_mean = self.inputs.reslice_which[1] > 0

        if isdefined(self.inputs.in_files):
            outputs['realignment_parameters'] = []
        for imgf in self.inputs.in_files:
            if isinstance(imgf, list):
                tmp_imgf = imgf[0]
            else:
                tmp_imgf = imgf
            outputs['realignment_parameters'].append(fname_presuffix(tmp_imgf,
                                                                     prefix='rp_',
                                                                     suffix='.txt',
                                                                     use_ext=False))
            if not isinstance(imgf, list) and func_is_3d(imgf):
                break

        if isinstance(self.inputs.in_files[0], list):
            first_image = self.inputs.in_files[0][0]
        else:
            first_image = self.inputs.in_files[0]

        if resliced_mean:
            outputs['mean_image'] = fname_presuffix(first_image, prefix='meanu')

        if resliced_all:
            outputs['realigned_unwarped_files'] = []
            for idx, imgf in enumerate(ensure_list(self.inputs.in_files)):
                realigned_run = []
                if isinstance(imgf, list):
                    for i, inner_imgf in enumerate(ensure_list(imgf)):
                        newfile = fname_presuffix(inner_imgf,
                                                  prefix=self.inputs.out_prefix)
                        realigned_run.append(newfile)
                else:
                    realigned_run = fname_presuffix(imgf,
                                                    prefix=self.inputs.out_prefix)
                outputs['realigned_unwarped_files'].append(realigned_run)
        return outputs


class CoregisterInputSpec(SPMCommandInputSpec):
    target = ImageFileSPM(
        exists=True,
        mandatory=True,
        field='ref',
        desc='reference file to register to',
        copyfile=False)
    source = InputMultiPath(
        ImageFileSPM(exists=True),
        field='source',
        desc='file to register to target',
        copyfile=True,
        mandatory=True)
    jobtype = traits.Enum(
        'estwrite',
        'estimate',
        'write',
        desc='one of: estimate, write, estwrite',
        usedefault=True)
    apply_to_files = InputMultiPath(
        File(exists=True),
        field='other',
        desc='files to apply transformation to',
        copyfile=True)
    cost_function = traits.Enum(
        'mi',
        'nmi',
        'ecc',
        'ncc',
        field='eoptions.cost_fun',
        desc="""cost function, one of:
                    'mi' - Mutual Information,
                    'nmi' - Normalised Mutual Information,
                    'ecc' - Entropy Correlation Coefficient,
                    'ncc' - Normalised Cross Correlation""")
    fwhm = traits.List(
        traits.Float(),
        minlen=2,
        maxlen=2,
        field='eoptions.fwhm',
        desc='gaussian smoothing kernel width (mm)')
    separation = traits.List(
        traits.Float(), field='eoptions.sep', desc='sampling separation in mm')
    tolerance = traits.List(
        traits.Float(),
        field='eoptions.tol',
        desc='acceptable tolerance for each of 12 params')
    write_interp = traits.Range(
        low=0,
        high=7,
        field='roptions.interp',
        desc=('degree of b-spline used for '
              'interpolation'))
    write_wrap = traits.List(
        traits.Int(),
        minlen=3,
        maxlen=3,
        field='roptions.wrap',
        desc=('Check if interpolation should wrap in '
              '[x,y,z]'))
    write_mask = traits.Bool(
        field='roptions.mask', desc='True/False mask output image')
    out_prefix = traits.String(
        'r',
        field='roptions.prefix',
        usedefault=True,
        desc='coregistered output prefix')


class CoregisterOutputSpec(TraitedSpec):
    coregistered_source = OutputMultiPath(
        File(exists=True), desc='Coregistered source files')
    coregistered_files = OutputMultiPath(
        File(exists=True), desc='Coregistered other files')


class Coregister(SPMCommand):
    """Use spm_coreg for estimating cross-modality rigid body alignment

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=39

    Examples
    --------

    >>> import nipype.interfaces.spm as spm
    >>> coreg = spm.Coregister()
    >>> coreg.inputs.target = 'functional.nii'
    >>> coreg.inputs.source = 'structural.nii'
    >>> coreg.run() # doctest: +SKIP

    """

    input_spec = CoregisterInputSpec
    output_spec = CoregisterOutputSpec
    _jobtype = 'spatial'
    _jobname = 'coreg'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if (opt == 'target'
                or (opt == 'source' and self.inputs.jobtype != "write")):
            return scans_for_fnames(ensure_list(val), keep4d=True)
        if opt == 'apply_to_files':
            return np.array(ensure_list(val), dtype=object)
        if opt == 'source' and self.inputs.jobtype == "write":
            if isdefined(self.inputs.apply_to_files):
                return scans_for_fnames(val + self.inputs.apply_to_files)
            else:
                return scans_for_fnames(val)
        return super(Coregister, self)._format_arg(opt, spec, val)

    def _parse_inputs(self):
        """validate spm coregister options if set to None ignore
        """
        if self.inputs.jobtype == "write":
            einputs = (super(Coregister, self)
                       ._parse_inputs(skip=('jobtype', 'apply_to_files')))
        else:
            einputs = super(Coregister, self)._parse_inputs(skip=('jobtype'))
        jobtype = self.inputs.jobtype
        return [{'%s' % (jobtype): einputs[0]}]

    def _list_outputs(self):
        outputs = self._outputs().get()

        if self.inputs.jobtype == "estimate":
            if isdefined(self.inputs.apply_to_files):
                outputs['coregistered_files'] = self.inputs.apply_to_files
            outputs['coregistered_source'] = self.inputs.source
        elif (self.inputs.jobtype == "write"
              or self.inputs.jobtype == "estwrite"):
            if isdefined(self.inputs.apply_to_files):
                outputs['coregistered_files'] = []
                for imgf in ensure_list(self.inputs.apply_to_files):
                    (outputs['coregistered_files'].append(
                        fname_presuffix(imgf, prefix=self.inputs.out_prefix)))

            outputs['coregistered_source'] = []
            for imgf in ensure_list(self.inputs.source):
                (outputs['coregistered_source'].append(
                    fname_presuffix(imgf, prefix=self.inputs.out_prefix)))

        return outputs


class NormalizeInputSpec(SPMCommandInputSpec):
    template = File(
        exists=True,
        field='eoptions.template',
        desc='template file to normalize to',
        mandatory=True,
        xor=['parameter_file'],
        copyfile=False)
    source = InputMultiPath(
        ImageFileSPM(exists=True),
        field='subj.source',
        xor=['parameter_file'],
        desc='file to normalize to template',
        mandatory=True,
        copyfile=True)
    jobtype = traits.Enum(
        'estwrite',
        'est',
        'write',
        usedefault=True,
        desc='Estimate, Write or do both')
    apply_to_files = InputMultiPath(
        traits.Either(File(exists=True), traits.List(File(exists=True))),
        field='subj.resample',
        desc='files to apply transformation to',
        copyfile=True)
    parameter_file = File(
        field='subj.matname',
        mandatory=True,
        xor=['source', 'template'],
        desc='normalization parameter file*_sn.mat',
        copyfile=False)
    source_weight = File(
        field='subj.wtsrc',
        desc='name of weighting image for source',
        copyfile=False)
    template_weight = File(
        field='eoptions.weight',
        desc='name of weighting image for template',
        copyfile=False)
    source_image_smoothing = traits.Float(
        field='eoptions.smosrc', desc='source smoothing')
    template_image_smoothing = traits.Float(
        field='eoptions.smoref', desc='template smoothing')
    affine_regularization_type = traits.Enum(
        'mni',
        'size',
        'none',
        field='eoptions.regtype',
        desc='mni, size, none')
    DCT_period_cutoff = traits.Float(
        field='eoptions.cutoff', desc='Cutoff of for DCT bases')
    nonlinear_iterations = traits.Int(
        field='eoptions.nits',
        desc=('Number of iterations of '
              'nonlinear warping'))
    nonlinear_regularization = traits.Float(
        field='eoptions.reg',
        desc=('the amount of the '
              'regularization for the '
              'nonlinear part of the '
              'normalization'))
    write_preserve = traits.Bool(
        field='roptions.preserve',
        desc='True/False warped images are modulated')
    write_bounding_box = traits.List(
        traits.List(traits.Float(), minlen=3, maxlen=3),
        field='roptions.bb',
        minlen=2,
        maxlen=2,
        desc='3x2-element list of lists')
    write_voxel_sizes = traits.List(
        traits.Float(),
        field='roptions.vox',
        minlen=3,
        maxlen=3,
        desc='3-element list')
    write_interp = traits.Range(
        low=0,
        high=7,
        field='roptions.interp',
        desc=('degree of b-spline used for '
              'interpolation'))
    write_wrap = traits.List(
        traits.Int(),
        field='roptions.wrap',
        desc=('Check if interpolation should wrap in '
              '[x,y,z] - list of bools'))
    out_prefix = traits.String(
        'w',
        field='roptions.prefix',
        usedefault=True,
        desc='normalized output prefix')


class NormalizeOutputSpec(TraitedSpec):
    normalization_parameters = OutputMultiPath(
        File(exists=True),
        desc=('MAT files containing '
              'the normalization '
              'parameters'))
    normalized_source = OutputMultiPath(
        File(exists=True), desc='Normalized source files')
    normalized_files = OutputMultiPath(
        File(exists=True), desc='Normalized other files')


class Normalize(SPMCommand):
    """use spm_normalise for warping an image to a template

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=203

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> norm = spm.Normalize()
    >>> norm.inputs.source = 'functional.nii'
    >>> norm.run() # doctest: +SKIP

    """

    input_spec = NormalizeInputSpec
    output_spec = NormalizeOutputSpec
    _jobtype = 'spatial'
    _jobname = 'normalise'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt == 'template':
            return scans_for_fname(ensure_list(val))
        if opt == 'source':
            return scans_for_fname(ensure_list(val))
        if opt == 'apply_to_files':
            return scans_for_fnames(ensure_list(val))
        if opt == 'parameter_file':
            return np.array([simplify_list(val)], dtype=object)
        if opt in ['write_wrap']:
            if len(val) != 3:
                raise ValueError('%s must have 3 elements' % opt)
        return super(Normalize, self)._format_arg(opt, spec, val)

    def _parse_inputs(self):
        """Validate spm normalize options if set to None ignore
        """
        einputs = super(
            Normalize, self)._parse_inputs(skip=('jobtype', 'apply_to_files'))
        if isdefined(self.inputs.apply_to_files):
            inputfiles = deepcopy(self.inputs.apply_to_files)
            if isdefined(self.inputs.source):
                inputfiles.extend(self.inputs.source)
            einputs[0]['subj']['resample'] = scans_for_fnames(inputfiles)
        jobtype = self.inputs.jobtype
        if jobtype in ['estwrite', 'write']:
            if not isdefined(self.inputs.apply_to_files):
                if isdefined(self.inputs.source):
                    einputs[0]['subj']['resample'] = scans_for_fname(
                        self.inputs.source)
        return [{'%s' % (jobtype): einputs[0]}]

    def _list_outputs(self):
        outputs = self._outputs().get()

        jobtype = self.inputs.jobtype
        if jobtype.startswith('est'):
            outputs['normalization_parameters'] = []
            for imgf in ensure_list(self.inputs.source):
                outputs['normalization_parameters'].append(
                    fname_presuffix(imgf, suffix='_sn.mat', use_ext=False))
            outputs['normalization_parameters'] = simplify_list(
                outputs['normalization_parameters'])

        if self.inputs.jobtype == "estimate":
            if isdefined(self.inputs.apply_to_files):
                outputs['normalized_files'] = self.inputs.apply_to_files
            outputs['normalized_source'] = self.inputs.source
        elif 'write' in self.inputs.jobtype:
            if (isdefined(self.inputs.write_preserve)
                    and self.inputs.write_preserve):
                prefixNorm = ''.join(['m', self.inputs.out_prefix])
            else:
                prefixNorm = self.inputs.out_prefix
            outputs['normalized_files'] = []
            if isdefined(self.inputs.apply_to_files):
                filelist = ensure_list(self.inputs.apply_to_files)
                for f in filelist:
                    if isinstance(f, list):
                        run = [
                            fname_presuffix(in_f, prefix=prefixNorm)
                            for in_f in f
                        ]
                    else:
                        run = [fname_presuffix(f, prefix=prefixNorm)]
                    outputs['normalized_files'].extend(run)
            if isdefined(self.inputs.source):
                outputs['normalized_source'] = []
                for imgf in ensure_list(self.inputs.source):
                    outputs['normalized_source'].append(
                        fname_presuffix(imgf, prefix=prefixNorm))

        return outputs


class Normalize12InputSpec(SPMCommandInputSpec):
    image_to_align = ImageFileSPM(
        exists=True,
        field='subj.vol',
        desc=('file to estimate normalization parameters '
              'with'),
        xor=['deformation_file'],
        mandatory=True,
        copyfile=True)
    apply_to_files = InputMultiPath(
        traits.Either(
            ImageFileSPM(exists=True), traits.List(ImageFileSPM(exists=True))),
        field='subj.resample',
        desc='files to apply transformation to',
        copyfile=True)
    deformation_file = ImageFileSPM(
        field='subj.def',
        mandatory=True,
        xor=['image_to_align', 'tpm'],
        copyfile=False,
        desc=('file y_*.nii containing 3 deformation '
              'fields for the deformation in x, y and z '
              'dimension'))
    jobtype = traits.Enum(
        'estwrite',
        'est',
        'write',
        usedefault=True,
        desc='Estimate, Write or do Both')
    bias_regularization = traits.Enum(
        0,
        0.00001,
        0.0001,
        0.001,
        0.01,
        0.1,
        1,
        10,
        field='eoptions.biasreg',
        desc='no(0) - extremely heavy (10)')
    bias_fwhm = traits.Enum(
        30,
        40,
        50,
        60,
        70,
        80,
        90,
        100,
        110,
        120,
        130,
        140,
        150,
        'Inf',
        field='eoptions.biasfwhm',
        desc='FWHM of Gaussian smoothness of bias')
    tpm = File(
        exists=True,
        field='eoptions.tpm',
        desc=('template in form of tissue probablitiy maps to '
              'normalize to'),
        xor=['deformation_file'],
        copyfile=False)
    affine_regularization_type = traits.Enum(
        'mni', 'size', 'none', field='eoptions.affreg', desc='mni, size, none')
    warping_regularization = traits.List(
        traits.Float(),
        field='eoptions.reg',
        minlen=5,
        maxlen=5,
        desc=('controls balance between '
              'parameters and data'))
    smoothness = traits.Float(
        field='eoptions.fwhm',
        desc=('value (in mm) to smooth the data before '
              'normalization'))
    sampling_distance = traits.Float(
        field='eoptions.samp',
        desc=('Sampling distance on data for '
              'parameter estimation'))
    write_bounding_box = traits.List(
        traits.List(traits.Float(), minlen=3, maxlen=3),
        field='woptions.bb',
        minlen=2,
        maxlen=2,
        desc=('3x2-element list of lists '
              'representing the bounding box '
              '(in mm) to be written'))
    write_voxel_sizes = traits.List(
        traits.Float(),
        field='woptions.vox',
        minlen=3,
        maxlen=3,
        desc=('3-element list representing the '
              'voxel sizes (in mm) of the written '
              'normalised images'))
    write_interp = traits.Range(
        low=0,
        high=7,
        field='woptions.interp',
        desc=('degree of b-spline used for '
              'interpolation'))
    out_prefix = traits.String(
        'w',
        field='woptions.prefix',
        usedefault=True,
        desc='Normalized output prefix')


class Normalize12OutputSpec(TraitedSpec):
    deformation_field = OutputMultiPath(
        File(exists=True),
        desc=('NIfTI file containing 3 '
              'deformation fields for the '
              'deformation in x, y and z '
              'dimension'))
    normalized_image = OutputMultiPath(
        File(exists=True),
        desc=('Normalized file that needed to '
              'be aligned'))
    normalized_files = OutputMultiPath(
        File(exists=True), desc='Normalized other files')


class Normalize12(SPMCommand):
    """uses SPM12's new Normalise routine for warping an image to a template.
    Spatial normalisation is now done via the segmentation routine (which was
    known as ``New Segment`` in SPM8). Note that the normalisation in SPM12
    is done towards a file containing multiple tissue probability maps, which
    was not the case in SPM8.

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=49

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> norm12 = spm.Normalize12()
    >>> norm12.inputs.image_to_align = 'structural.nii'
    >>> norm12.inputs.apply_to_files = 'functional.nii'
    >>> norm12.run() # doctest: +SKIP

    """

    input_spec = Normalize12InputSpec
    output_spec = Normalize12OutputSpec
    _jobtype = 'spatial'
    _jobname = 'normalise'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt == 'tpm':
            return scans_for_fname(ensure_list(val))
        if opt == 'image_to_align':
            return scans_for_fname(ensure_list(val))
        if opt == 'apply_to_files':
            return scans_for_fnames(ensure_list(val))
        if opt == 'deformation_file':
            return np.array([simplify_list(val)], dtype=object)
        if opt in ['nonlinear_regularization']:
            if len(val) != 5:
                raise ValueError('%s must have 5 elements' % opt)
        return super(Normalize12, self)._format_arg(opt, spec, val)

    def _parse_inputs(self, skip=()):
        """validate spm normalize options if set to None ignore
        """
        einputs = super(
            Normalize12,
            self)._parse_inputs(skip=('jobtype', 'apply_to_files'))
        if isdefined(self.inputs.apply_to_files):
            inputfiles = deepcopy(self.inputs.apply_to_files)
            if isdefined(self.inputs.image_to_align):
                inputfiles.extend([self.inputs.image_to_align])
            einputs[0]['subj']['resample'] = scans_for_fnames(inputfiles)
        jobtype = self.inputs.jobtype
        if jobtype in ['estwrite', 'write']:
            if not isdefined(self.inputs.apply_to_files):
                if isdefined(self.inputs.image_to_align):
                    einputs[0]['subj']['resample'] = scans_for_fname(
                        self.inputs.image_to_align)
        return [{'%s' % (jobtype): einputs[0]}]

    def _list_outputs(self):
        outputs = self._outputs().get()

        jobtype = self.inputs.jobtype
        if jobtype.startswith('est'):
            outputs['deformation_field'] = []
            for imgf in ensure_list(self.inputs.image_to_align):
                outputs['deformation_field'].append(
                    fname_presuffix(imgf, prefix='y_'))
            outputs['deformation_field'] = simplify_list(
                outputs['deformation_field'])

        if self.inputs.jobtype == "estimate":
            if isdefined(self.inputs.apply_to_files):
                outputs['normalized_files'] = self.inputs.apply_to_files
            outputs['normalized_image'] = fname_presuffix(
                self.inputs.image_to_align, prefix='w')
        elif 'write' in self.inputs.jobtype:
            outputs['normalized_files'] = []
            if isdefined(self.inputs.apply_to_files):
                filelist = ensure_list(self.inputs.apply_to_files)
                for f in filelist:
                    if isinstance(f, list):
                        run = [fname_presuffix(in_f, prefix='w') for in_f in f]
                    else:
                        run = [fname_presuffix(f, prefix='w')]
                    outputs['normalized_files'].extend(run)
            if isdefined(self.inputs.image_to_align):
                outputs['normalized_image'] = fname_presuffix(
                    self.inputs.image_to_align, prefix='w')

        return outputs


class SegmentInputSpec(SPMCommandInputSpec):
    data = InputMultiPath(
        ImageFileSPM(exists=True),
        field='data',
        desc='one scan per subject',
        copyfile=False,
        mandatory=True)
    gm_output_type = traits.List(
        traits.Bool(),
        minlen=3,
        maxlen=3,
        field='output.GM',
        desc=
        """Options to produce grey matter images: c1*.img, wc1*.img and mwc1*.img.
            None: [False,False,False],
            Native Space: [False,False,True],
            Unmodulated Normalised: [False,True,False],
            Modulated Normalised: [True,False,False],
            Native + Unmodulated Normalised: [False,True,True],
            Native + Modulated Normalised: [True,False,True],
            Native + Modulated + Unmodulated: [True,True,True],
            Modulated + Unmodulated Normalised: [True,True,False]""")
    wm_output_type = traits.List(
        traits.Bool(),
        minlen=3,
        maxlen=3,
        field='output.WM',
        desc="""
            Options to produce white matter images: c2*.img, wc2*.img and mwc2*.img.
            None: [False,False,False],
            Native Space: [False,False,True],
            Unmodulated Normalised: [False,True,False],
            Modulated Normalised: [True,False,False],
            Native + Unmodulated Normalised: [False,True,True],
            Native + Modulated Normalised: [True,False,True],
            Native + Modulated + Unmodulated: [True,True,True],
            Modulated + Unmodulated Normalised: [True,True,False]""")
    csf_output_type = traits.List(
        traits.Bool(),
        minlen=3,
        maxlen=3,
        field='output.CSF',
        desc="""
            Options to produce CSF images: c3*.img, wc3*.img and mwc3*.img.
            None: [False,False,False],
            Native Space: [False,False,True],
            Unmodulated Normalised: [False,True,False],
            Modulated Normalised: [True,False,False],
            Native + Unmodulated Normalised: [False,True,True],
            Native + Modulated Normalised: [True,False,True],
            Native + Modulated + Unmodulated: [True,True,True],
            Modulated + Unmodulated Normalised: [True,True,False]""")
    save_bias_corrected = traits.Bool(
        field='output.biascor',
        desc=('True/False produce a bias '
              'corrected image'))
    clean_masks = traits.Enum(
        'no',
        'light',
        'thorough',
        field='output.cleanup',
        desc=("clean using estimated brain mask "
              "('no','light','thorough')"))
    tissue_prob_maps = traits.List(
        File(exists=True),
        field='opts.tpm',
        desc=('list of gray, white & csf prob. '
              '(opt,)'))
    gaussians_per_class = traits.List(
        traits.Int(),
        field='opts.ngaus',
        desc=('num Gaussians capture intensity '
              'distribution'))
    affine_regularization = traits.Enum(
        'mni',
        'eastern',
        'subj',
        'none',
        '',
        field='opts.regtype',
        desc=('Possible options: "mni", '
              '"eastern", "subj", "none" '
              '(no reguralisation), "" '
              '(no affine registration)'))
    warping_regularization = traits.Float(
        field='opts.warpreg',
        desc=('Controls balance between '
              'parameters and data'))
    warp_frequency_cutoff = traits.Float(
        field='opts.warpco', desc='Cutoff of DCT bases')
    bias_regularization = traits.Enum(
        0,
        0.00001,
        0.0001,
        0.001,
        0.01,
        0.1,
        1,
        10,
        field='opts.biasreg',
        desc='no(0) - extremely heavy (10)')
    bias_fwhm = traits.Enum(
        30,
        40,
        50,
        60,
        70,
        80,
        90,
        100,
        110,
        120,
        130,
        'Inf',
        field='opts.biasfwhm',
        desc='FWHM of Gaussian smoothness of bias')
    sampling_distance = traits.Float(
        field='opts.samp',
        desc=('Sampling distance on data for '
              'parameter estimation'))
    mask_image = File(
        exists=True,
        field='opts.msk',
        desc='Binary image to restrict parameter estimation ')


class SegmentOutputSpec(TraitedSpec):
    native_gm_image = File(desc='native space grey probability map')
    normalized_gm_image = File(desc='normalized grey probability map', )
    modulated_gm_image = File(
        desc=('modulated, normalized grey '
              'probability map'))
    native_wm_image = File(desc='native space white probability map')
    normalized_wm_image = File(desc='normalized white probability map')
    modulated_wm_image = File(
        desc=('modulated, normalized white '
              'probability map'))
    native_csf_image = File(desc='native space csf probability map')
    normalized_csf_image = File(desc='normalized csf probability map')
    modulated_csf_image = File(
        desc=('modulated, normalized csf '
              'probability map'))
    modulated_input_image = File(
        deprecated='0.10',
        new_name='bias_corrected_image',
        desc='bias-corrected version of input image')
    bias_corrected_image = File(desc='bias-corrected version of input image')
    transformation_mat = File(exists=True, desc='Normalization transformation')
    inverse_transformation_mat = File(
        exists=True, desc='Inverse normalization info')


class Segment(SPMCommand):
    """use spm_segment to separate structural images into different
    tissue classes.

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=209

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> seg = spm.Segment()
    >>> seg.inputs.data = 'structural.nii'
    >>> seg.run() # doctest: +SKIP

    """

    input_spec = SegmentInputSpec
    output_spec = SegmentOutputSpec

    def __init__(self, **inputs):
        _local_version = SPMCommand().version
        if _local_version and '12.' in _local_version:
            self._jobtype = 'tools'
            self._jobname = 'oldseg'
        else:
            self._jobtype = 'spatial'
            self._jobname = 'preproc'

        SPMCommand.__init__(self, **inputs)

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        clean_masks_dict = {'no': 0, 'light': 1, 'thorough': 2}

        if opt in ['data', 'tissue_prob_maps']:
            if isinstance(val, list):
                return scans_for_fnames(val)
            else:
                return scans_for_fname(val)
        if 'output_type' in opt:
            return [int(v) for v in val]
        if opt == 'mask_image':
            return scans_for_fname(val)
        if opt == 'clean_masks':
            return clean_masks_dict[val]
        return super(Segment, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        f = self.inputs.data[0]

        for tidx, tissue in enumerate(['gm', 'wm', 'csf']):
            outtype = '%s_output_type' % tissue
            if isdefined(getattr(self.inputs, outtype)):
                for idx, (image, prefix) in enumerate([('modulated', 'mw'),
                                                       ('normalized',
                                                        'w'), ('native', '')]):
                    if getattr(self.inputs, outtype)[idx]:
                        outfield = '%s_%s_image' % (image, tissue)
                        outputs[outfield] = fname_presuffix(
                            f, prefix='%sc%d' % (prefix, tidx + 1))
        if (isdefined(self.inputs.save_bias_corrected)
                and self.inputs.save_bias_corrected):
            outputs['bias_corrected_image'] = fname_presuffix(f, prefix='m')
        t_mat = fname_presuffix(f, suffix='_seg_sn.mat', use_ext=False)
        outputs['transformation_mat'] = t_mat
        invt_mat = fname_presuffix(f, suffix='_seg_inv_sn.mat', use_ext=False)
        outputs['inverse_transformation_mat'] = invt_mat
        return outputs


class NewSegmentInputSpec(SPMCommandInputSpec):
    channel_files = InputMultiPath(
        ImageFileSPM(exists=True),
        mandatory=True,
        desc="A list of files to be segmented",
        field='channel',
        copyfile=False)
    channel_info = traits.Tuple(
        traits.Float(),
        traits.Float(),
        traits.Tuple(traits.Bool, traits.Bool),
        desc="""A tuple with the following fields:
            - bias reguralisation (0-10)
            - FWHM of Gaussian smoothness of bias
            - which maps to save (Field, Corrected) - a tuple of two boolean values""",
        field='channel')
    tissues = traits.List(
        traits.Tuple(
            traits.Tuple(ImageFileSPM(exists=True), traits.Int()),
            traits.Int(), traits.Tuple(traits.Bool, traits.Bool),
            traits.Tuple(traits.Bool, traits.Bool)),
        desc="""A list of tuples (one per tissue) with the following fields:
            - tissue probability map (4D), 1-based index to frame
            - number of gaussians
            - which maps to save [Native, DARTEL] - a tuple of two boolean values
            - which maps to save [Unmodulated, Modulated] - a tuple of two boolean values""",
        field='tissue')
    affine_regularization = traits.Enum(
        'mni',
        'eastern',
        'subj',
        'none',
        field='warp.affreg',
        desc='mni, eastern, subj, none ')
    warping_regularization = traits.Either(
        traits.List(traits.Float(), minlen=5, maxlen=5),
        traits.Float(),
        field='warp.reg',
        desc=('Warping regularization '
              'parameter(s). Accepts float '
              'or list of floats (the '
              'latter is required by '
              'SPM12)'))
    sampling_distance = traits.Float(
        field='warp.samp',
        desc=('Sampling distance on data for '
              'parameter estimation'))
    write_deformation_fields = traits.List(
        traits.Bool(),
        minlen=2,
        maxlen=2,
        field='warp.write',
        desc=("Which deformation fields to "
              "write:[Inverse, Forward]"))


class NewSegmentOutputSpec(TraitedSpec):
    native_class_images = traits.List(
        traits.List(File(exists=True)), desc='native space probability maps')
    dartel_input_images = traits.List(
        traits.List(File(exists=True)), desc='dartel imported class images')
    normalized_class_images = traits.List(
        traits.List(File(exists=True)), desc='normalized class images')
    modulated_class_images = traits.List(
        traits.List(File(exists=True)),
        desc=('modulated+normalized class '
              'images'))
    transformation_mat = OutputMultiPath(
        File(exists=True), desc='Normalization transformation')
    bias_corrected_images = OutputMultiPath(
        File(exists=True), desc='bias corrected images')
    bias_field_images = OutputMultiPath(
        File(exists=True), desc='bias field images')
    forward_deformation_field = OutputMultiPath(File(exists=True))
    inverse_deformation_field = OutputMultiPath(File(exists=True))


class NewSegment(SPMCommand):
    """Use spm_preproc8 (New Segment) to separate structural images into
    different tissue classes. Supports multiple modalities.

    NOTE: This interface currently supports single channel input only

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=43

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> seg = spm.NewSegment()
    >>> seg.inputs.channel_files = 'structural.nii'
    >>> seg.inputs.channel_info = (0.0001, 60, (True, True))
    >>> seg.run() # doctest: +SKIP

    For VBM pre-processing [http://www.fil.ion.ucl.ac.uk/~john/misc/VBMclass10.pdf],
    TPM.nii should be replaced by /path/to/spm8/toolbox/Seg/TPM.nii

    >>> seg = NewSegment()
    >>> seg.inputs.channel_files = 'structural.nii'
    >>> tissue1 = (('TPM.nii', 1), 2, (True,True), (False, False))
    >>> tissue2 = (('TPM.nii', 2), 2, (True,True), (False, False))
    >>> tissue3 = (('TPM.nii', 3), 2, (True,False), (False, False))
    >>> tissue4 = (('TPM.nii', 4), 2, (False,False), (False, False))
    >>> tissue5 = (('TPM.nii', 5), 2, (False,False), (False, False))
    >>> seg.inputs.tissues = [tissue1, tissue2, tissue3, tissue4, tissue5]
    >>> seg.run() # doctest: +SKIP

    """

    input_spec = NewSegmentInputSpec
    output_spec = NewSegmentOutputSpec

    def __init__(self, **inputs):
        _local_version = SPMCommand().version
        if _local_version and '12.' in _local_version:
            self._jobtype = 'spatial'
            self._jobname = 'preproc'
        else:
            self._jobtype = 'tools'
            self._jobname = 'preproc8'

        SPMCommand.__init__(self, **inputs)

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """

        if opt in ['channel_files', 'channel_info']:
            # structure have to be recreated because of some weird traits error
            new_channel = {}
            new_channel['vols'] = scans_for_fnames(self.inputs.channel_files)
            if isdefined(self.inputs.channel_info):
                info = self.inputs.channel_info
                new_channel['biasreg'] = info[0]
                new_channel['biasfwhm'] = info[1]
                new_channel['write'] = [int(info[2][0]), int(info[2][1])]
            return [new_channel]
        elif opt == 'tissues':
            new_tissues = []
            for tissue in val:
                new_tissue = {}
                new_tissue['tpm'] = np.array(
                    [','.join([tissue[0][0], str(tissue[0][1])])],
                    dtype=object)
                new_tissue['ngaus'] = tissue[1]
                new_tissue['native'] = [int(tissue[2][0]), int(tissue[2][1])]
                new_tissue['warped'] = [int(tissue[3][0]), int(tissue[3][1])]
                new_tissues.append(new_tissue)
            return new_tissues
        elif opt == 'write_deformation_fields':
            return super(NewSegment, self)._format_arg(
                opt, spec, [int(val[0]), int(val[1])])
        else:
            return super(NewSegment, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['native_class_images'] = []
        outputs['dartel_input_images'] = []
        outputs['normalized_class_images'] = []
        outputs['modulated_class_images'] = []
        outputs['transformation_mat'] = []
        outputs['bias_corrected_images'] = []
        outputs['bias_field_images'] = []
        outputs['inverse_deformation_field'] = []
        outputs['forward_deformation_field'] = []

        n_classes = 5
        if isdefined(self.inputs.tissues):
            n_classes = len(self.inputs.tissues)
        for i in range(n_classes):
            outputs['native_class_images'].append([])
            outputs['dartel_input_images'].append([])
            outputs['normalized_class_images'].append([])
            outputs['modulated_class_images'].append([])

        for filename in self.inputs.channel_files:
            pth, base, ext = split_filename(filename)
            if isdefined(self.inputs.tissues):
                for i, tissue in enumerate(self.inputs.tissues):
                    if tissue[2][0]:
                        outputs['native_class_images'][i].append(
                            os.path.join(pth, "c%d%s.nii" % (i + 1, base)))
                    if tissue[2][1]:
                        outputs['dartel_input_images'][i].append(
                            os.path.join(pth, "rc%d%s.nii" % (i + 1, base)))
                    if tissue[3][0]:
                        outputs['normalized_class_images'][i].append(
                            os.path.join(pth, "wc%d%s.nii" % (i + 1, base)))
                    if tissue[3][1]:
                        outputs['modulated_class_images'][i].append(
                            os.path.join(pth, "mwc%d%s.nii" % (i + 1, base)))
            else:
                for i in range(n_classes):
                    outputs['native_class_images'][i].append(
                        os.path.join(pth, "c%d%s.nii" % (i + 1, base)))
            outputs['transformation_mat'].append(
                os.path.join(pth, "%s_seg8.mat" % base))

            if isdefined(self.inputs.write_deformation_fields):
                if self.inputs.write_deformation_fields[0]:
                    outputs['inverse_deformation_field'].append(
                        os.path.join(pth, "iy_%s.nii" % base))
                if self.inputs.write_deformation_fields[1]:
                    outputs['forward_deformation_field'].append(
                        os.path.join(pth, "y_%s.nii" % base))

            if isdefined(self.inputs.channel_info):
                if self.inputs.channel_info[2][0]:
                    outputs['bias_field_images'].append(
                        os.path.join(pth, "BiasField_%s.nii" % (base)))
                if self.inputs.channel_info[2][1]:
                    outputs['bias_corrected_images'].append(
                        os.path.join(pth, "m%s.nii" % (base)))
        return outputs


class SmoothInputSpec(SPMCommandInputSpec):
    in_files = InputMultiPath(
        ImageFileSPM(exists=True),
        field='data',
        desc='list of files to smooth',
        mandatory=True,
        copyfile=False)
    fwhm = traits.Either(
        traits.List(traits.Float(), minlen=3, maxlen=3),
        traits.Float(),
        field='fwhm',
        desc='3-list of fwhm for each dimension')
    data_type = traits.Int(
        field='dtype', desc='Data type of the output images')
    implicit_masking = traits.Bool(
        field='im', desc=('A mask implied by a particular'
                          'voxel value'))
    out_prefix = traits.String(
        's', field='prefix', usedefault=True, desc='smoothed output prefix')


class SmoothOutputSpec(TraitedSpec):
    smoothed_files = OutputMultiPath(File(exists=True), desc='smoothed files')


class Smooth(SPMCommand):
    """Use spm_smooth for 3D Gaussian smoothing of image volumes.

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=55

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> smooth = spm.Smooth()
    >>> smooth.inputs.in_files = 'functional.nii'
    >>> smooth.inputs.fwhm = [4, 4, 4]
    >>> smooth.run() # doctest: +SKIP
    """

    input_spec = SmoothInputSpec
    output_spec = SmoothOutputSpec
    _jobtype = 'spatial'
    _jobname = 'smooth'

    def _format_arg(self, opt, spec, val):
        if opt in ['in_files']:
            return scans_for_fnames(ensure_list(val))
        if opt == 'fwhm':
            if not isinstance(val, list):
                return [val, val, val]
            if isinstance(val, list):
                if len(val) == 1:
                    return [val[0], val[0], val[0]]
                else:
                    return val

        return super(Smooth, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['smoothed_files'] = []

        for imgf in ensure_list(self.inputs.in_files):
            outputs['smoothed_files'].append(
                fname_presuffix(imgf, prefix=self.inputs.out_prefix))
        return outputs


class DARTELInputSpec(SPMCommandInputSpec):
    image_files = traits.List(
        traits.List(ImageFileSPM(exists=True)),
        desc="A list of files to be segmented",
        field='warp.images',
        copyfile=False,
        mandatory=True)
    template_prefix = traits.Str(
        'Template',
        usedefault=True,
        field='warp.settings.template',
        desc='Prefix for template')
    regularization_form = traits.Enum(
        'Linear',
        'Membrane',
        'Bending',
        field='warp.settings.rform',
        desc=('Form of regularization energy '
              'term'))
    iteration_parameters = traits.List(
        traits.Tuple(
            traits.Range(1, 10),
            traits.Tuple(traits.Float, traits.Float, traits.Float),
            traits.Enum(1, 2, 4, 8, 16, 32, 64, 128, 256, 512),
            traits.Enum(0, 0.5, 1, 2, 4, 8, 16, 32)),
        minlen=3,
        maxlen=12,
        field='warp.settings.param',
        desc="""List of tuples for each iteration
                                       - Inner iterations
                                       - Regularization parameters
                                       - Time points for deformation model
                                       - smoothing parameter
                                       """)
    optimization_parameters = traits.Tuple(
        traits.Float,
        traits.Range(1, 8),
        traits.Range(1, 8),
        field='warp.settings.optim',
        desc="""
                                           Optimization settings a tuple
                                           - LM regularization
                                           - cycles of multigrid solver
                                           - relaxation iterations
                                           """)


class DARTELOutputSpec(TraitedSpec):
    final_template_file = File(exists=True, desc='final DARTEL template')
    template_files = traits.List(
        File(exists=True),
        desc=('Templates from different stages of '
              'iteration'))
    dartel_flow_fields = traits.List(
        File(exists=True), desc='DARTEL flow fields')


class DARTEL(SPMCommand):
    """Use spm DARTEL to create a template and flow fields

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=185

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> dartel = spm.DARTEL()
    >>> dartel.inputs.image_files = [['rc1s1.nii','rc1s2.nii'],['rc2s1.nii', 'rc2s2.nii']]
    >>> dartel.run() # doctest: +SKIP

    """

    input_spec = DARTELInputSpec
    output_spec = DARTELOutputSpec
    _jobtype = 'tools'
    _jobname = 'dartel'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """

        if opt in ['image_files']:
            return scans_for_fnames(val, keep4d=True, separate_sessions=True)
        elif opt == 'regularization_form':
            mapper = {'Linear': 0, 'Membrane': 1, 'Bending': 2}
            return mapper[val]
        elif opt == 'iteration_parameters':
            params = []
            for param in val:
                new_param = {}
                new_param['its'] = param[0]
                new_param['rparam'] = list(param[1])
                new_param['K'] = param[2]
                new_param['slam'] = param[3]
                params.append(new_param)
            return params
        elif opt == 'optimization_parameters':
            new_param = {}
            new_param['lmreg'] = val[0]
            new_param['cyc'] = val[1]
            new_param['its'] = val[2]
            return [new_param]
        else:
            return super(DARTEL, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['template_files'] = []
        for i in range(6):
            outputs['template_files'].append(
                os.path.realpath('%s_%d.nii' % (self.inputs.template_prefix,
                                                i + 1)))
        outputs['final_template_file'] = os.path.realpath(
            '%s_6.nii' % self.inputs.template_prefix)
        outputs['dartel_flow_fields'] = []
        for filename in self.inputs.image_files[0]:
            pth, base, ext = split_filename(filename)
            outputs['dartel_flow_fields'].append(
                os.path.realpath('u_%s_%s%s' %
                                 (base, self.inputs.template_prefix, ext)))
        return outputs


class DARTELNorm2MNIInputSpec(SPMCommandInputSpec):
    template_file = ImageFileSPM(
        exists=True,
        copyfile=False,
        mandatory=True,
        desc="DARTEL template",
        field='mni_norm.template')
    flowfield_files = InputMultiPath(
        ImageFileSPM(exists=True),
        mandatory=True,
        desc="DARTEL flow fields u_rc1*",
        field='mni_norm.data.subjs.flowfields')
    apply_to_files = InputMultiPath(
        ImageFileSPM(exists=True),
        desc="Files to apply the transform to",
        field='mni_norm.data.subjs.images',
        mandatory=True,
        copyfile=False)
    voxel_size = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        desc="Voxel sizes for output file",
        field='mni_norm.vox')
    bounding_box = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        traits.Float,
        traits.Float,
        traits.Float,
        desc="Voxel sizes for output file",
        field='mni_norm.bb')
    modulate = traits.Bool(
        field='mni_norm.preserve',
        desc=("Modulate out images - no modulation "
              "preserves concentrations"))
    fwhm = traits.Either(
        traits.List(traits.Float(), minlen=3, maxlen=3),
        traits.Float(),
        field='mni_norm.fwhm',
        desc='3-list of fwhm for each dimension')


class DARTELNorm2MNIOutputSpec(TraitedSpec):
    normalized_files = OutputMultiPath(
        File(exists=True), desc='Normalized files in MNI space')
    normalization_parameter_file = File(
        exists=True, desc=('Transform parameters to MNI '
                           'space'))


class DARTELNorm2MNI(SPMCommand):
    """Use spm DARTEL to normalize data to MNI space

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=188

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> nm = spm.DARTELNorm2MNI()
    >>> nm.inputs.template_file = 'Template_6.nii'
    >>> nm.inputs.flowfield_files = ['u_rc1s1_Template.nii', 'u_rc1s3_Template.nii']
    >>> nm.inputs.apply_to_files = ['c1s1.nii', 'c1s3.nii']
    >>> nm.inputs.modulate = True
    >>> nm.run() # doctest: +SKIP

    """

    input_spec = DARTELNorm2MNIInputSpec
    output_spec = DARTELNorm2MNIOutputSpec
    _jobtype = 'tools'
    _jobname = 'dartel'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ['template_file']:
            return np.array([val], dtype=object)
        elif opt in ['flowfield_files']:
            return scans_for_fnames(val, keep4d=True)
        elif opt in ['apply_to_files']:
            return scans_for_fnames(val, keep4d=True, separate_sessions=True)
        elif opt == 'voxel_size':
            return list(val)
        elif opt == 'bounding_box':
            return list(val)
        elif opt == 'fwhm':
            if isinstance(val, list):
                return val
            else:
                return [val, val, val]
        else:
            return super(DARTELNorm2MNI, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        pth, base, ext = split_filename(self.inputs.template_file)
        outputs['normalization_parameter_file'] = os.path.realpath(
            base + '_2mni.mat')
        outputs['normalized_files'] = []
        prefix = "w"
        if isdefined(self.inputs.modulate) and self.inputs.modulate:
            prefix = 'm' + prefix
        if not isdefined(self.inputs.fwhm) or self.inputs.fwhm > 0:
            prefix = 's' + prefix
        for filename in self.inputs.apply_to_files:
            pth, base, ext = split_filename(filename)
            outputs['normalized_files'].append(
                os.path.realpath('%s%s%s' % (prefix, base, ext)))

        return outputs


class CreateWarpedInputSpec(SPMCommandInputSpec):
    image_files = InputMultiPath(
        ImageFileSPM(exists=True),
        mandatory=True,
        desc="A list of files to be warped",
        field='crt_warped.images',
        copyfile=False)
    flowfield_files = InputMultiPath(
        ImageFileSPM(exists=True),
        copyfile=False,
        desc="DARTEL flow fields u_rc1*",
        field='crt_warped.flowfields',
        mandatory=True)
    iterations = traits.Range(
        low=0,
        high=9,
        desc=("The number of iterations: log2(number of "
              "time steps)"),
        field='crt_warped.K')
    interp = traits.Range(
        low=0,
        high=7,
        field='crt_warped.interp',
        desc='degree of b-spline used for interpolation')
    modulate = traits.Bool(
        field='crt_warped.jactransf', desc="Modulate images")


class CreateWarpedOutputSpec(TraitedSpec):
    warped_files = traits.List(File(exists=True, desc='final warped files'))


class CreateWarped(SPMCommand):
    """Apply a flow field estimated by DARTEL to create warped images

    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=190

    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> create_warped = spm.CreateWarped()
    >>> create_warped.inputs.image_files = ['rc1s1.nii', 'rc1s2.nii']
    >>> create_warped.inputs.flowfield_files = ['u_rc1s1_Template.nii', 'u_rc1s2_Template.nii']
    >>> create_warped.run() # doctest: +SKIP

    """

    input_spec = CreateWarpedInputSpec
    output_spec = CreateWarpedOutputSpec
    _jobtype = 'tools'
    _jobname = 'dartel'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """

        if opt in ['image_files']:
            return scans_for_fnames(val, keep4d=True, separate_sessions=True)
        if opt in ['flowfield_files']:
            return scans_for_fnames(val, keep4d=True)
        else:
            return super(CreateWarped, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['warped_files'] = []
        for filename in self.inputs.image_files:
            pth, base, ext = split_filename(filename)
            if isdefined(self.inputs.modulate) and self.inputs.modulate:
                outputs['warped_files'].append(
                    os.path.realpath('mw%s%s' % (base, ext)))
            else:
                outputs['warped_files'].append(
                    os.path.realpath('w%s%s' % (base, ext)))
        return outputs


class ApplyDeformationFieldInputSpec(SPMCommandInputSpec):
    in_files = InputMultiPath(
        ImageFileSPM(exists=True), mandatory=True, field='fnames')
    deformation_field = File(exists=True, mandatory=True, field='comp{1}.def')
    reference_volume = ImageFileSPM(
        exists=True, mandatory=True, field='comp{2}.id.space')
    interp = traits.Range(
        low=0,
        high=7,
        field='interp',
        desc='degree of b-spline used for interpolation')


class ApplyDeformationFieldOutputSpec(TraitedSpec):
    out_files = OutputMultiPath(File(exists=True))


class ApplyDeformations(SPMCommand):
    input_spec = ApplyDeformationFieldInputSpec
    output_spec = ApplyDeformationFieldOutputSpec

    _jobtype = 'util'
    _jobname = 'defs'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ['deformation_field', 'reference_volume']:
            val = [val]

        if opt in ['deformation_field']:
            return scans_for_fnames(val, keep4d=True, separate_sessions=False)
        if opt in ['in_files', 'reference_volume']:
            return scans_for_fnames(val, keep4d=False, separate_sessions=False)

        else:
            return super(ApplyDeformations, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_files'] = []
        for filename in self.inputs.in_files:
            _, fname = os.path.split(filename)
            outputs['out_files'].append(os.path.realpath('w%s' % fname))
        return outputs


class VBMSegmentInputSpec(SPMCommandInputSpec):

    in_files = InputMultiPath(
        ImageFileSPM(exists=True),
        desc="A list of files to be segmented",
        field='estwrite.data',
        copyfile=False,
        mandatory=True)

    tissues = ImageFileSPM(
        exists=True, field='estwrite.tpm', desc='tissue probability map')
    gaussians_per_class = traits.Tuple(
        (2, 2, 2, 3, 4, 2),
        *([traits.Int()] * 6),
        usedefault=True,
        desc='number of gaussians for each tissue class')
    bias_regularization = traits.Enum(
        0.0001, (0, 0.00001, 0.0001, 0.001, 0.01, 0.1, 1, 10),
        field='estwrite.opts.biasreg',
        usedefault=True,
        desc='no(0) - extremely heavy (10)')

    bias_fwhm = traits.Enum(
        60, (30, 40, 50, 60, 70, 80, 90, 100, 110, 120, 130, 'Inf'),
        field='estwrite.opts.biasfwhm',
        usedefault=True,
        desc='FWHM of Gaussian smoothness of bias')
    sampling_distance = traits.Float(
        3,
        usedefault=True,
        field='estwrite.opts.samp',
        desc='Sampling distance on data for parameter estimation')
    warping_regularization = traits.Float(
        4,
        usedefault=True,
        field='estwrite.opts.warpreg',
        desc='Controls balance between parameters and data')

    spatial_normalization = traits.Enum(
        'high',
        'low',
        usedefault=True,
    )
    dartel_template = ImageFileSPM(
        exists=True, field='estwrite.extopts.dartelwarp.normhigh.darteltpm')
    use_sanlm_denoising_filter = traits.Range(
        0,
        2,
        2,
        usedefault=True,
        field='estwrite.extopts.sanlm',
        desc="0=No denoising, 1=denoising,2=denoising multi-threaded")
    mrf_weighting = traits.Float(
        0.15, usedefault=True, field='estwrite.extopts.mrf')
    cleanup_partitions = traits.Int(
        1,
        usedefault=True,
        field='estwrite.extopts.cleanup',
        desc="0=None,1=light,2=thorough")
    display_results = traits.Bool(
        True, usedefault=True, field='estwrite.extopts.print')

    gm_native = traits.Bool(
        False,
        usedefault=True,
        field='estwrite.output.GM.native',
    )
    gm_normalized = traits.Bool(
        False,
        usedefault=True,
        field='estwrite.output.GM.warped',
    )
    gm_modulated_normalized = traits.Range(
        0,
        2,
        2,
        usedefault=True,
        field='estwrite.output.GM.modulated',
        desc='0=none,1=affine+non-linear(SPM8 default),2=non-linear only')
    gm_dartel = traits.Range(
        0,
        2,
        0,
        usedefault=True,
        field='estwrite.output.GM.dartel',
        desc="0=None,1=rigid(SPM8 default),2=affine")

    wm_native = traits.Bool(
        False,
        usedefault=True,
        field='estwrite.output.WM.native',
    )
    wm_normalized = traits.Bool(
        False,
        usedefault=True,
        field='estwrite.output.WM.warped',
    )
    wm_modulated_normalized = traits.Range(
        0,
        2,
        2,
        usedefault=True,
        field='estwrite.output.WM.modulated',
        desc='0=none,1=affine+non-linear(SPM8 default),2=non-linear only')
    wm_dartel = traits.Range(
        0,
        2,
        0,
        usedefault=True,
        field='estwrite.output.WM.dartel',
        desc="0=None,1=rigid(SPM8 default),2=affine")

    csf_native = traits.Bool(
        False,
        usedefault=True,
        field='estwrite.output.CSF.native',
    )
    csf_normalized = traits.Bool(
        False,
        usedefault=True,
        field='estwrite.output.CSF.warped',
    )
    csf_modulated_normalized = traits.Range(
        0,
        2,
        2,
        usedefault=True,
        field='estwrite.output.CSF.modulated',
        desc='0=none,1=affine+non-linear(SPM8 default),2=non-linear only')
    csf_dartel = traits.Range(
        0,
        2,
        0,
        usedefault=True,
        field='estwrite.output.CSF.dartel',
        desc="0=None,1=rigid(SPM8 default),2=affine")

    bias_corrected_native = traits.Bool(
        False,
        usedefault=True,
        field='estwrite.output.bias.native',
    )
    bias_corrected_normalized = traits.Bool(
        True,
        usedefault=True,
        field='estwrite.output.bias.warped',
    )
    bias_corrected_affine = traits.Bool(
        False,
        usedefault=True,
        field='estwrite.output.bias.affine',
    )

    pve_label_native = traits.Bool(
        False, usedefault=True, field='estwrite.output.label.native')
    pve_label_normalized = traits.Bool(
        False, usedefault=True, field='estwrite.output.label.warped')
    pve_label_dartel = traits.Range(
        0,
        2,
        0,
        usedefault=True,
        field='estwrite.output.label.dartel',
        desc="0=None,1=rigid(SPM8 default),2=affine")

    jacobian_determinant = traits.Bool(
        False, usedefault=True, field='estwrite.jacobian.warped')

    deformation_field = traits.Tuple(
        (0, 0),
        traits.Bool,
        traits.Bool,
        usedefault=True,
        field='estwrite.output.warps',
        desc='forward and inverse field')


class VBMSegmentOuputSpec(TraitedSpec):

    native_class_images = traits.List(
        traits.List(File(exists=True)), desc='native space probability maps')
    dartel_input_images = traits.List(
        traits.List(File(exists=True)), desc='dartel imported class images')
    normalized_class_images = traits.List(
        traits.List(File(exists=True)), desc='normalized class images')
    modulated_class_images = traits.List(
        traits.List(File(exists=True)),
        desc=('modulated+normalized class '
              'images'))
    transformation_mat = OutputMultiPath(
        File(exists=True), desc='Normalization transformation')

    bias_corrected_images = OutputMultiPath(
        File(exists=True), desc='bias corrected images')
    normalized_bias_corrected_images = OutputMultiPath(
        File(exists=True), desc='bias corrected images')

    pve_label_native_images = OutputMultiPath(File(exists=True))
    pve_label_normalized_images = OutputMultiPath(File(exists=True))
    pve_label_registered_images = OutputMultiPath(File(exists=True))

    forward_deformation_field = OutputMultiPath(File(exists=True))
    inverse_deformation_field = OutputMultiPath(File(exists=True))

    jacobian_determinant_images = OutputMultiPath(File(exists=True))


class VBMSegment(SPMCommand):
    """Use VBM8 toolbox to separate structural images into different
    tissue classes.

    Example
    -------
    >>> import nipype.interfaces.spm as spm
    >>> seg = spm.VBMSegment()
    >>> seg.inputs.tissues = 'TPM.nii'
    >>> seg.inputs.dartel_template = 'Template_1_IXI550_MNI152.nii'
    >>> seg.inputs.bias_corrected_native = True
    >>> seg.inputs.gm_native = True
    >>> seg.inputs.wm_native = True
    >>> seg.inputs.csf_native = True
    >>> seg.inputs.pve_label_native = True
    >>> seg.inputs.deformation_field = (True, False)
    >>> seg.run() # doctest: +SKIP
    """

    input_spec = VBMSegmentInputSpec
    output_spec = VBMSegmentOuputSpec

    _jobtype = 'tools'
    _jobname = 'vbm8'

    def _list_outputs(self):
        outputs = self._outputs().get()

        do_dartel = self.inputs.spatial_normalization
        dartel_px = ''
        if do_dartel:
            dartel_px = 'r'

        outputs['native_class_images'] = [[], [], []]
        outputs['dartel_input_images'] = [[], [], []]
        outputs['normalized_class_images'] = [[], [], []]
        outputs['modulated_class_images'] = [[], [], []]

        outputs['transformation_mat'] = []

        outputs['bias_corrected_images'] = []
        outputs['normalized_bias_corrected_images'] = []

        outputs['inverse_deformation_field'] = []
        outputs['forward_deformation_field'] = []
        outputs['jacobian_determinant_images'] = []

        outputs['pve_label_native_images'] = []
        outputs['pve_label_normalized_images'] = []
        outputs['pve_label_registered_images'] = []

        for filename in self.inputs.in_files:
            pth, base, ext = split_filename(filename)

            outputs['transformation_mat'].append(
                os.path.join(pth, "%s_seg8.mat" % base))

            for i, tis in enumerate(['gm', 'wm', 'csf']):
                # native space

                if getattr(self.inputs, '%s_native' % tis):
                    outputs['native_class_images'][i].append(
                        os.path.join(pth, "p%d%s.nii" % (i + 1, base)))
                if getattr(self.inputs, '%s_dartel' % tis) == 1:
                    outputs['dartel_input_images'][i].append(
                        os.path.join(pth, "rp%d%s.nii" % (i + 1, base)))
                elif getattr(self.inputs, '%s_dartel' % tis) == 2:
                    outputs['dartel_input_images'][i].append(
                        os.path.join(pth, "rp%d%s_affine.nii" % (i + 1, base)))

            # normalized space
                if getattr(self.inputs, '%s_normalized' % tis):
                    outputs['normalized_class_images'][i].append(
                        os.path.join(pth, "w%sp%d%s.nii" % (dartel_px, i + 1,
                                                            base)))

                if getattr(self.inputs, '%s_modulated_normalized' % tis) == 1:
                    outputs['modulated_class_images'][i].append(
                        os.path.join(pth, "mw%sp%d%s.nii" % (dartel_px, i + 1,
                                                             base)))
                elif getattr(self.inputs,
                             '%s_modulated_normalized' % tis) == 2:
                    outputs['normalized_class_images'][i].append(
                        os.path.join(pth, "m0w%sp%d%s.nii" % (dartel_px, i + 1,
                                                              base)))

            if self.inputs.pve_label_native:
                outputs['pve_label_native_images'].append(
                    os.path.join(pth, "p0%s.nii" % (base)))
            if self.inputs.pve_label_normalized:
                outputs['pve_label_normalized_images'].append(
                    os.path.join(pth, "w%sp0%s.nii" % (dartel_px, base)))
            if self.inputs.pve_label_dartel == 1:
                outputs['pve_label_registered_images'].append(
                    os.path.join(pth, "rp0%s.nii" % (base)))
            elif self.inputs.pve_label_dartel == 2:
                outputs['pve_label_registered_images'].append(
                    os.path.join(pth, "rp0%s_affine.nii" % (base)))

            if self.inputs.bias_corrected_native:
                outputs['bias_corrected_images'].append(
                    os.path.join(pth, "m%s.nii" % (base)))
            if self.inputs.bias_corrected_normalized:
                outputs['normalized_bias_corrected_images'].append(
                    os.path.join(pth, "wm%s%s.nii" % (dartel_px, base)))

            if self.inputs.deformation_field[0]:
                outputs['forward_deformation_field'].append(
                    os.path.join(pth, "y_%s%s.nii" % (dartel_px, base)))
            if self.inputs.deformation_field[1]:
                outputs['inverse_deformation_field'].append(
                    os.path.join(pth, "iy_%s%s.nii" % (dartel_px, base)))

            if self.inputs.jacobian_determinant and do_dartel:
                outputs['jacobian_determinant_images'].append(
                    os.path.join(pth, "jac_wrp1%s.nii" % (base)))
        return outputs

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm
        """
        if opt in ['in_files']:
            return scans_for_fnames(val, keep4d=True)
        elif opt in ['spatial_normalization']:
            if val == 'low':
                return {'normlow': []}
        elif opt in ['dartel_template']:
            return np.array([val], dtype=object)
        elif opt in ['deformation_field']:
            return super(VBMSegment, self)._format_arg(
                opt, spec, [int(val[0]), int(val[1])])
        else:
            return super(VBMSegment, self)._format_arg(opt, spec, val)

    def _parse_inputs(self):
        if self.inputs.spatial_normalization == 'low':
            einputs = super(VBMSegment, self)._parse_inputs(
                skip=('spatial_normalization', 'dartel_template'))
            einputs[0]['estwrite']['extopts']['dartelwarp'] = {'normlow': 1}
            return einputs
        else:
            return super(VBMSegment,
                         self)._parse_inputs(skip=('spatial_normalization'))



# -*- coding: utf-8 -*-
"""
Handles custom functions used in Function interface. Future imports
are avoided to keep namespace as clear as possible.
"""
from builtins import next, str
from future.utils import raise_from
import inspect
from textwrap import dedent


def getsource(function):
    """Returns the source code of a function"""
    return dedent(inspect.getsource(function))


def create_function_from_source(function_source, imports=None):
    """Return a function object from a function source

    Parameters
    ----------
    function_source : unicode string
        unicode string defining a function
    imports : list of strings
        list of import statements in string form that allow the function
        to be executed in an otherwise empty namespace
    """
    ns = {}
    import_keys = []

    try:
        if imports is not None:
            for statement in imports:
                exec(statement, ns)
            import_keys = list(ns.keys())
        exec(function_source, ns)

    except Exception as e:
        msg = 'Error executing function\n{}\n'.format(function_source)
        msg += ("Functions in connection strings have to be standalone. "
                "They cannot be declared either interactively or inside "
                "another function or inline in the connect string. Any "
                "imports should be done inside the function.")
        raise_from(RuntimeError(msg), e)
    ns_funcs = list(set(ns) - set(import_keys + ['__builtins__']))
    assert len(ns_funcs) == 1, "Function or inputs are ill-defined"
    func = ns[ns_funcs[0]]
    return func



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from .specialized import (ACPCTransform, FiducialRegistration,
                          VBRAINSDemonWarp, BRAINSDemonWarp)
from .brainsresample import BRAINSResample
from .brainsfit import BRAINSFit



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import MRIsCalc


def test_MRIsCalc_inputs():
    input_map = dict(
        action=dict(
            argstr='%s',
            mandatory=True,
            position=-2,
        ),
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file1=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=-3,
        ),
        in_file2=dict(
            argstr='%s',
            extensions=None,
            position=-1,
            xor=['in_float', 'in_int'],
        ),
        in_float=dict(
            argstr='%f',
            position=-1,
            xor=['in_file2', 'in_int'],
        ),
        in_int=dict(
            argstr='%d',
            position=-1,
            xor=['in_file2', 'in_float'],
        ),
        out_file=dict(
            argstr='-o %s',
            extensions=None,
            mandatory=True,
        ),
        subjects_dir=dict(),
    )
    inputs = MRIsCalc.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_MRIsCalc_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = MRIsCalc.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..minc import BBox


def test_BBox_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        format_minccrop=dict(argstr='-minccrop', ),
        format_mincresample=dict(argstr='-mincresample', ),
        format_mincreshape=dict(argstr='-mincreshape', ),
        input_file=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        one_line=dict(
            argstr='-one_line',
            xor=('one_line', 'two_lines'),
        ),
        out_file=dict(
            argstr='> %s',
            extensions=None,
            genfile=True,
            position=-1,
        ),
        output_file=dict(
            extensions=None,
            hash_files=False,
            keep_extension=False,
            name_source=['input_file'],
            name_template='%s_bbox.txt',
            position=-1,
        ),
        threshold=dict(argstr='-threshold', ),
        two_lines=dict(
            argstr='-two_lines',
            xor=('one_line', 'two_lines'),
        ),
    )
    inputs = BBox.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_BBox_outputs():
    output_map = dict(output_file=dict(extensions=None, ), )
    outputs = BBox.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import CommandLine, CommandLineInputSpec, SEMLikeCommandLine, TraitedSpec, File, Directory, traits, isdefined, InputMultiPath, OutputMultiPath
import os


class ThresholdScalarVolumeInputSpec(CommandLineInputSpec):
    InputVolume = File(
        position=-2, desc="Input volume", exists=True, argstr="%s")
    OutputVolume = traits.Either(
        traits.Bool,
        File(),
        position=-1,
        hash_files=False,
        desc="Thresholded input volume",
        argstr="%s")
    threshold = traits.Int(desc="Threshold value", argstr="--threshold %d")
    lower = traits.Int(desc="Lower threshold value", argstr="--lower %d")
    upper = traits.Int(desc="Upper threshold value", argstr="--upper %d")
    outsidevalue = traits.Int(
        desc=
        "Set the voxels to this value if they fall outside the threshold range",
        argstr="--outsidevalue %d")
    thresholdtype = traits.Enum(
        "Below",
        "Above",
        "Outside",
        desc=
        "What kind of threshold to perform. If Outside is selected, uses Upper and Lower values. If Below is selected, uses the ThresholdValue, if Above is selected, uses the ThresholdValue.",
        argstr="--thresholdtype %s")


class ThresholdScalarVolumeOutputSpec(TraitedSpec):
    OutputVolume = File(
        position=-1, desc="Thresholded input volume", exists=True)


class ThresholdScalarVolume(SEMLikeCommandLine):
    """title: Threshold Scalar Volume

category: Filtering

description: <p>Threshold an image.</p><p>Set image values to a user-specified outside value if they are below, above, or between simple threshold values.</p><p>ThresholdAbove: The values greater than or equal to the threshold value are set to OutsideValue.</p><p>ThresholdBelow: The values less than or equal to the threshold value are set to OutsideValue.</p><p>ThresholdOutside: The values outside the range Lower-Upper are set to OutsideValue.</p><p>Although all image types are supported on input, only signed types are produced.</p><p>

version: 0.1.0.$Revision: 2104 $(alpha)

documentation-url: http://wiki.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/Threshold

contributor: Nicole Aucoin (SPL, BWH), Ron Kikinis (SPL, BWH)

acknowledgements: This work is part of the National Alliance for Medical Image Computing (NAMIC), funded by the National Institutes of Health through the NIH Roadmap for Medical Research, Grant U54 EB005149.

"""

    input_spec = ThresholdScalarVolumeInputSpec
    output_spec = ThresholdScalarVolumeOutputSpec
    _cmd = "ThresholdScalarVolume "
    _outputs_filenames = {'OutputVolume': 'OutputVolume.nii'}



# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os.path as op
import nibabel as nb
import numpy as np
from distutils.version import LooseVersion

from ...utils import NUMPY_MMAP

from ... import logging
from ..base import (traits, TraitedSpec, File, isdefined)
from .base import (HAVE_DIPY, dipy_version, dipy_to_nipype_interface,
                   get_dipy_workflows, DipyBaseInterface)

IFLOGGER = logging.getLogger('nipype.interface')

if HAVE_DIPY and LooseVersion(dipy_version()) >= LooseVersion('0.15'):
    from dipy.workflows import denoise, mask

    l_wkflw = get_dipy_workflows(denoise) + get_dipy_workflows(mask)
    for name, obj in l_wkflw:
        new_name = name.replace('Flow', '')
        globals()[new_name] = dipy_to_nipype_interface(new_name, obj)
    del l_wkflw

else:
    IFLOGGER.info("We advise you to upgrade DIPY version. This upgrade will"
                  " open access to more function")


class ResampleInputSpec(TraitedSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        desc='The input 4D diffusion-weighted image file')
    vox_size = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        desc=('specify the new voxel zooms. If no vox_size'
              ' is set, then isotropic regridding will '
              'be performed, with spacing equal to the '
              'smallest current zoom.'))
    interp = traits.Int(
        1,
        mandatory=True,
        usedefault=True,
        desc=('order of the interpolator (0 = nearest, 1 = linear, etc.'))


class ResampleOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class Resample(DipyBaseInterface):
    """
    An interface to reslicing diffusion datasets.
    See
    http://nipy.org/dipy/examples_built/reslice_datasets.html#example-reslice-datasets.

    Example
    -------

    >>> import nipype.interfaces.dipy as dipy
    >>> reslice = dipy.Resample()
    >>> reslice.inputs.in_file = 'diffusion.nii'
    >>> reslice.run() # doctest: +SKIP
    """
    input_spec = ResampleInputSpec
    output_spec = ResampleOutputSpec

    def _run_interface(self, runtime):
        order = self.inputs.interp
        vox_size = None

        if isdefined(self.inputs.vox_size):
            vox_size = self.inputs.vox_size

        out_file = op.abspath(self._gen_outfilename())
        resample_proxy(
            self.inputs.in_file,
            order=order,
            new_zooms=vox_size,
            out_file=out_file)

        IFLOGGER.info('Resliced image saved as %s', out_file)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = op.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        fname, fext = op.splitext(op.basename(self.inputs.in_file))
        if fext == '.gz':
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext
        return op.abspath('%s_reslice%s' % (fname, fext))


class DenoiseInputSpec(TraitedSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        desc='The input 4D diffusion-weighted image file')
    in_mask = File(exists=True, desc='brain mask')
    noise_model = traits.Enum(
        'rician',
        'gaussian',
        mandatory=True,
        usedefault=True,
        desc=('noise distribution model'))
    signal_mask = File(
        desc=('mask in which the mean signal '
              'will be computed'),
        exists=True)
    noise_mask = File(
        desc=('mask in which the standard deviation of noise '
              'will be computed'),
        exists=True)
    patch_radius = traits.Int(1, usedefault=True, desc='patch radius')
    block_radius = traits.Int(5, usedefault=True, desc='block_radius')
    snr = traits.Float(desc='manually set an SNR')


class DenoiseOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class Denoise(DipyBaseInterface):
    """
    An interface to denoising diffusion datasets [Coupe2008]_.
    See
    http://nipy.org/dipy/examples_built/denoise_nlmeans.html#example-denoise-nlmeans.

    .. [Coupe2008] Coupe P et al., `An Optimized Blockwise Non Local Means
      Denoising Filter for 3D Magnetic Resonance Images
      <https://doi.org/10.1109%2FTMI.2007.906087>`_,
      IEEE Transactions on Medical Imaging, 27(4):425-441, 2008.


    Example
    -------

    >>> import nipype.interfaces.dipy as dipy
    >>> denoise = dipy.Denoise()
    >>> denoise.inputs.in_file = 'diffusion.nii'
    >>> denoise.run() # doctest: +SKIP
    """
    input_spec = DenoiseInputSpec
    output_spec = DenoiseOutputSpec

    def _run_interface(self, runtime):
        out_file = op.abspath(self._gen_outfilename())

        settings = dict(
            mask=None, rician=(self.inputs.noise_model == 'rician'))

        if isdefined(self.inputs.in_mask):
            settings['mask'] = nb.load(self.inputs.in_mask).get_data()

        if isdefined(self.inputs.patch_radius):
            settings['patch_radius'] = self.inputs.patch_radius

        if isdefined(self.inputs.block_radius):
            settings['block_radius'] = self.inputs.block_radius

        snr = None
        if isdefined(self.inputs.snr):
            snr = self.inputs.snr

        signal_mask = None
        if isdefined(self.inputs.signal_mask):
            signal_mask = nb.load(self.inputs.signal_mask).get_data()
        noise_mask = None
        if isdefined(self.inputs.noise_mask):
            noise_mask = nb.load(self.inputs.noise_mask).get_data()

        _, s = nlmeans_proxy(
            self.inputs.in_file,
            settings,
            snr=snr,
            smask=signal_mask,
            nmask=noise_mask,
            out_file=out_file)
        IFLOGGER.info('Denoised image saved as %s, estimated SNR=%s', out_file,
                      str(s))
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = op.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        fname, fext = op.splitext(op.basename(self.inputs.in_file))
        if fext == '.gz':
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext
        return op.abspath('%s_denoise%s' % (fname, fext))


def resample_proxy(in_file, order=3, new_zooms=None, out_file=None):
    """
    Performs regridding of an image to set isotropic voxel sizes using dipy.
    """
    from dipy.align.reslice import reslice

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == '.gz':
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext
        out_file = op.abspath('./%s_reslice%s' % (fname, fext))

    img = nb.load(in_file, mmap=NUMPY_MMAP)
    hdr = img.header.copy()
    data = img.get_data().astype(np.float32)
    affine = img.affine
    im_zooms = hdr.get_zooms()[:3]

    if new_zooms is None:
        minzoom = np.array(im_zooms).min()
        new_zooms = tuple(np.ones((3, )) * minzoom)

    if np.all(im_zooms == new_zooms):
        return in_file

    data2, affine2 = reslice(data, affine, im_zooms, new_zooms, order=order)
    tmp_zooms = np.array(hdr.get_zooms())
    tmp_zooms[:3] = new_zooms[0]
    hdr.set_zooms(tuple(tmp_zooms))
    hdr.set_data_shape(data2.shape)
    hdr.set_xyzt_units('mm')
    nb.Nifti1Image(data2.astype(hdr.get_data_dtype()), affine2,
                   hdr).to_filename(out_file)
    return out_file, new_zooms


def nlmeans_proxy(in_file,
                  settings,
                  snr=None,
                  smask=None,
                  nmask=None,
                  out_file=None):
    """
    Uses non-local means to denoise 4D datasets
    """
    from dipy.denoise.nlmeans import nlmeans
    from scipy.ndimage.morphology import binary_erosion
    from scipy import ndimage

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == '.gz':
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext
        out_file = op.abspath('./%s_denoise%s' % (fname, fext))

    img = nb.load(in_file, mmap=NUMPY_MMAP)
    hdr = img.header
    data = img.get_data()
    aff = img.affine

    if data.ndim < 4:
        data = data[..., np.newaxis]

    data = np.nan_to_num(data)

    if data.max() < 1.0e-4:
        raise RuntimeError('There is no signal in the image')

    df = 1.0
    if data.max() < 1000.0:
        df = 1000. / data.max()
        data *= df

    b0 = data[..., 0]

    if smask is None:
        smask = np.zeros_like(b0)
        smask[b0 > np.percentile(b0, 85.)] = 1

    smask = binary_erosion(
        smask.astype(np.uint8), iterations=2).astype(np.uint8)

    if nmask is None:
        nmask = np.ones_like(b0, dtype=np.uint8)
        bmask = settings['mask']
        if bmask is None:
            bmask = np.zeros_like(b0)
            bmask[b0 > np.percentile(b0[b0 > 0], 10)] = 1
            label_im, nb_labels = ndimage.label(bmask)
            sizes = ndimage.sum(bmask, label_im, range(nb_labels + 1))
            maxidx = np.argmax(sizes)
            bmask = np.zeros_like(b0, dtype=np.uint8)
            bmask[label_im == maxidx] = 1
        nmask[bmask > 0] = 0
    else:
        nmask = np.squeeze(nmask)
        nmask[nmask > 0.0] = 1
        nmask[nmask < 1] = 0
        nmask = nmask.astype(bool)

    nmask = binary_erosion(nmask, iterations=1).astype(np.uint8)

    den = np.zeros_like(data)

    est_snr = True
    if snr is not None:
        snr = [snr] * data.shape[-1]
        est_snr = False
    else:
        snr = []

    for i in range(data.shape[-1]):
        d = data[..., i]
        if est_snr:
            s = np.mean(d[smask > 0])
            n = np.std(d[nmask > 0])
            snr.append(s / n)

        den[..., i] = nlmeans(d, snr[i], **settings)

    den = np.squeeze(den)
    den /= df

    nb.Nifti1Image(den.astype(hdr.get_data_dtype()), aff,
                   hdr).to_filename(out_file)
    return out_file, snr



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import Bandpass


def test_Bandpass_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        automask=dict(argstr='-automask', ),
        blur=dict(argstr='-blur %f', ),
        despike=dict(argstr='-despike', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        highpass=dict(
            argstr='%f',
            mandatory=True,
            position=-3,
        ),
        in_file=dict(
            argstr='%s',
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=-1,
        ),
        localPV=dict(argstr='-localPV %f', ),
        lowpass=dict(
            argstr='%f',
            mandatory=True,
            position=-2,
        ),
        mask=dict(
            argstr='-mask %s',
            extensions=None,
            position=2,
        ),
        nfft=dict(argstr='-nfft %d', ),
        no_detrend=dict(argstr='-nodetrend', ),
        normalize=dict(argstr='-norm', ),
        notrans=dict(argstr='-notrans', ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        orthogonalize_dset=dict(
            argstr='-dsort %s',
            extensions=None,
        ),
        orthogonalize_file=dict(argstr='-ort %s', ),
        out_file=dict(
            argstr='-prefix %s',
            extensions=None,
            name_source='in_file',
            name_template='%s_bp',
            position=1,
        ),
        outputtype=dict(),
        tr=dict(argstr='-dt %f', ),
    )
    inputs = Bandpass.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Bandpass_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = Bandpass.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import LTAConvert


def test_LTAConvert_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_fsl=dict(
            argstr='--infsl %s',
            extensions=None,
            mandatory=True,
            xor=('in_lta', 'in_fsl', 'in_mni', 'in_reg', 'in_niftyreg',
                 'in_itk'),
        ),
        in_itk=dict(
            argstr='--initk %s',
            extensions=None,
            mandatory=True,
            xor=('in_lta', 'in_fsl', 'in_mni', 'in_reg', 'in_niftyreg',
                 'in_itk'),
        ),
        in_lta=dict(
            argstr='--inlta %s',
            mandatory=True,
            xor=('in_lta', 'in_fsl', 'in_mni', 'in_reg', 'in_niftyreg',
                 'in_itk'),
        ),
        in_mni=dict(
            argstr='--inmni %s',
            extensions=None,
            mandatory=True,
            xor=('in_lta', 'in_fsl', 'in_mni', 'in_reg', 'in_niftyreg',
                 'in_itk'),
        ),
        in_niftyreg=dict(
            argstr='--inniftyreg %s',
            extensions=None,
            mandatory=True,
            xor=('in_lta', 'in_fsl', 'in_mni', 'in_reg', 'in_niftyreg',
                 'in_itk'),
        ),
        in_reg=dict(
            argstr='--inreg %s',
            extensions=None,
            mandatory=True,
            xor=('in_lta', 'in_fsl', 'in_mni', 'in_reg', 'in_niftyreg',
                 'in_itk'),
        ),
        invert=dict(argstr='--invert', ),
        ltavox2vox=dict(
            argstr='--ltavox2vox',
            requires=['out_lta'],
        ),
        out_fsl=dict(argstr='--outfsl %s', ),
        out_itk=dict(argstr='--outitk %s', ),
        out_lta=dict(argstr='--outlta %s', ),
        out_mni=dict(argstr='--outmni %s', ),
        out_reg=dict(argstr='--outreg %s', ),
        source_file=dict(
            argstr='--src %s',
            extensions=None,
        ),
        target_conform=dict(argstr='--trgconform', ),
        target_file=dict(
            argstr='--trg %s',
            extensions=None,
        ),
    )
    inputs = LTAConvert.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_LTAConvert_outputs():
    output_map = dict(
        out_fsl=dict(extensions=None, ),
        out_itk=dict(extensions=None, ),
        out_lta=dict(extensions=None, ),
        out_mni=dict(extensions=None, ),
        out_reg=dict(extensions=None, ),
    )
    outputs = LTAConvert.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..n4itkbiasfieldcorrection import N4ITKBiasFieldCorrection


def test_N4ITKBiasFieldCorrection_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        bsplineorder=dict(argstr='--bsplineorder %d', ),
        convergencethreshold=dict(argstr='--convergencethreshold %f', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        histogramsharpening=dict(
            argstr='--histogramsharpening %s',
            sep=',',
        ),
        inputimage=dict(
            argstr='--inputimage %s',
            extensions=None,
        ),
        iterations=dict(
            argstr='--iterations %s',
            sep=',',
        ),
        maskimage=dict(
            argstr='--maskimage %s',
            extensions=None,
        ),
        meshresolution=dict(
            argstr='--meshresolution %s',
            sep=',',
        ),
        outputbiasfield=dict(
            argstr='--outputbiasfield %s',
            hash_files=False,
        ),
        outputimage=dict(
            argstr='--outputimage %s',
            hash_files=False,
        ),
        shrinkfactor=dict(argstr='--shrinkfactor %d', ),
        splinedistance=dict(argstr='--splinedistance %f', ),
        weightimage=dict(
            argstr='--weightimage %s',
            extensions=None,
        ),
    )
    inputs = N4ITKBiasFieldCorrection.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_N4ITKBiasFieldCorrection_outputs():
    output_map = dict(
        outputbiasfield=dict(extensions=None, ),
        outputimage=dict(extensions=None, ),
    )
    outputs = N4ITKBiasFieldCorrection.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from builtins import str, open, bytes
# This tool exports a Nipype interface in the Boutiques
# (https://github.com/boutiques) JSON format. Boutiques tools
# can be imported in CBRAIN (https://github.com/aces/cbrain)
# among other platforms.
#
# Limitations:
#   * Optional outputs, i.e. outputs that not always produced, may not be
#     detected. They will, however, still be listed with a placeholder for
#     the path template (either a value key or the output ID) that should
#     be verified and corrected.
#   * Still need to add some fields to the descriptor manually, e.g. url,
#     descriptor-url, path-template-stripped-extensions, etc.

import os
import sys
import simplejson as json

from ..scripts.instance import import_module


def generate_boutiques_descriptor(
        module, interface_name, container_image, container_type,
        container_index=None, verbose=False, save=False, save_path=None,
        author=None, ignore_inputs=None, tags=None):
    '''
    Returns a JSON string containing a JSON Boutiques description of a
    Nipype interface.
    Arguments:
    * module: module where the Nipype interface is declared.
    * interface_name: name of Nipype interface.
    * container_image: name of the container image where the tool is installed
    * container_type: type of container image (Docker or Singularity)
    * container_index: optional index where the image is available
    * verbose: print information messages
    * save: True if you want to save descriptor to a file
    * save_path: file path for the saved descriptor (defaults to name of the
      interface in current directory)
    * author: author of the tool (required for publishing)
    * ignore_inputs: list of interface inputs to not include in the descriptor
    * tags: JSON object containing tags to include in the descriptor,
      e.g. "{\"key1\": \"value1\"}" (note: the tags 'domain:neuroinformatics'
      and 'interface-type:nipype' are included by default)
    '''

    if not module:
        raise Exception("Undefined module.")

    # Retrieves Nipype interface
    if isinstance(module, (str, bytes)):
        import_module(module)
        module_name = str(module)
        module = sys.modules[module]
    else:
        module_name = str(module.__name__)

    interface = getattr(module, interface_name)()
    inputs = interface.input_spec()
    outputs = interface.output_spec()

    # Tool description
    tool_desc = {}
    tool_desc['name'] = interface_name
    tool_desc[
        'command-line'] = interface_name + " "
    tool_desc['author'] = "Nipype (interface)"
    if author is not None:
        tool_desc['author'] = tool_desc['author'] + ", " + author + " (tool)"
    tool_desc['description'] = (interface_name +
                                ", as implemented in Nipype (module: " +
                                module_name + ", interface: " +
                                interface_name + ").")
    tool_desc['inputs'] = []
    tool_desc['output-files'] = []
    tool_desc['groups'] = []
    tool_desc['tool-version'] = interface.version \
        if interface.version is not None else "1.0.0"
    tool_desc['schema-version'] = '0.5'
    if container_image:
        tool_desc['container-image'] = {}
        tool_desc['container-image']['image'] = container_image
        tool_desc['container-image']['type'] = container_type
        if container_index:
            tool_desc['container-image']['index'] = container_index

    # Generates tool inputs
    for name, spec in sorted(interface.inputs.traits(transient=None).items()):
        # Skip ignored inputs
        if ignore_inputs is not None and name in ignore_inputs:
            continue
        # If spec has a name source, this means it actually represents an
        # output, so create a Boutiques output from it
        elif spec.name_source and spec.name_template:
            tool_desc['output-files']\
                .append(get_boutiques_output_from_inp(inputs, spec, name))
        else:
            inp = get_boutiques_input(inputs, interface, name, spec, verbose)
            # Handle compound inputs (inputs that can be of multiple types
            # and are mutually exclusive)
            if isinstance(inp, list):
                mutex_group_members = []
                tool_desc['command-line'] += inp[0]['value-key'] + " "
                for i in inp:
                    tool_desc['inputs'].append(i)
                    mutex_group_members.append(i['id'])
                    if verbose:
                        print("-> Adding input " + i['name'])
                # Put inputs into a mutually exclusive group
                tool_desc['groups'].append({'id': inp[0]['id'] + "_group",
                                            'name': inp[0]['name'] + " group",
                                            'members': mutex_group_members,
                                            'mutually-exclusive': True})
            else:
                tool_desc['inputs'].append(inp)
                tool_desc['command-line'] += inp['value-key'] + " "
                if verbose:
                    print("-> Adding input " + inp['name'])

    # Generates input groups
    tool_desc['groups'] +=\
        get_boutiques_groups(interface.inputs.traits(transient=None).items())
    if len(tool_desc['groups']) == 0:
        del tool_desc['groups']

    # Generates tool outputs
    generate_tool_outputs(outputs, interface, tool_desc, verbose, True)

    # Generate outputs with various different inputs to try to generate
    # as many output values as possible
    custom_inputs = generate_custom_inputs(tool_desc['inputs'])

    for input_dict in custom_inputs:
        interface = getattr(module, interface_name)(**input_dict)
        outputs = interface.output_spec()
        generate_tool_outputs(outputs, interface, tool_desc, verbose, False)

    # Fill in all missing output paths
    for output in tool_desc['output-files']:
        if output['path-template'] == "":
            fill_in_missing_output_path(output, output['name'],
                                        tool_desc['inputs'])

    # Add tags
    desc_tags = {
        'domain': 'neuroinformatics',
        'source': 'nipype-interface'
    }

    if tags is not None:
        tags_dict = json.loads(tags)
        for k, v in tags_dict.items():
            if k in desc_tags:
                if not isinstance(desc_tags[k], list):
                    desc_tags[k] = [desc_tags[k]]
                desc_tags[k].append(v)
            else:
                desc_tags[k] = v

    tool_desc['tags'] = desc_tags

    # Check for positional arguments and reorder command line args if necessary
    tool_desc['command-line'] = reorder_cmd_line_args(
        tool_desc['command-line'], interface, ignore_inputs)

    # Remove the extra space at the end of the command line
    tool_desc['command-line'] = tool_desc['command-line'].strip()

    # Save descriptor to a file
    if save:
        path = save_path or os.path.join(os.getcwd(), interface_name + '.json')
        with open(path, 'w') as outfile:
            json.dump(tool_desc, outfile, indent=4, separators=(',', ': '))
        if verbose:
            print("-> Descriptor saved to file " + outfile.name)

    print("NOTE: Descriptors produced by this script may not entirely conform "
          "to the Nipype interface specs. Please check that the descriptor is "
          "correct before using it.")
    return json.dumps(tool_desc, indent=4, separators=(',', ': '))


def generate_tool_outputs(outputs, interface, tool_desc, verbose, first_run):
    for name, spec in sorted(outputs.traits(transient=None).items()):
        output = get_boutiques_output(outputs, name, spec, interface,
                                      tool_desc['inputs'])
        # If this is the first time we are generating outputs, add the full
        # output to the descriptor. Otherwise, find the existing output and
        # update its path template if it's still undefined.
        if first_run:
            tool_desc['output-files'].append(output)
            if output.get('value-key'):
                tool_desc['command-line'] += output['value-key'] + " "
            if verbose:
                print("-> Adding output " + output['name'])
        else:
            for existing_output in tool_desc['output-files']:
                if (output['id'] == existing_output['id'] and
                        existing_output['path-template'] == ""):
                    existing_output['path-template'] = output['path-template']
                    break
            if (output.get('value-key') and
                    output['value-key'] not in tool_desc['command-line']):
                tool_desc['command-line'] += output['value-key'] + " "

    if len(tool_desc['output-files']) == 0:
        raise Exception("Tool has no output.")


def get_boutiques_input(inputs, interface, input_name, spec, verbose,
                        handler=None, input_number=None):
    """
    Returns a dictionary containing the Boutiques input corresponding
    to a Nipype input.

    Args:
      * inputs: inputs of the Nipype interface.
      * interface: Nipype interface.
      * input_name: name of the Nipype input.
      * spec: Nipype input spec.
      * verbose: print information messages.
      * handler: used when handling compound inputs, which don't have their
        own input spec
      * input_number: used when handling compound inputs to assign each a
        unique ID

    Assumes that:
      * Input names are unique.
    """
    inp = {}

    # No need to append a number to the first of a list of compound inputs
    if input_number:
        inp['id'] = input_name + "_" + str(input_number + 1)
    else:
        inp['id'] = input_name

    inp['name'] = input_name.replace('_', ' ').capitalize()

    if handler is None:
        trait_handler = spec.handler
    else:
        trait_handler = handler

    # Figure out the input type from its handler type
    handler_type = type(trait_handler).__name__

    # Deal with compound traits
    if handler_type == "TraitCompound":
        input_list = []
        # Recursively create an input for each trait
        for i in range(0, len(trait_handler.handlers)):
            inp = get_boutiques_input(inputs, interface, input_name, spec,
                                      verbose, trait_handler.handlers[i], i)
            inp['optional'] = True
            input_list.append(inp)
        return input_list

    if handler_type == "File" or handler_type == "Directory":
        inp['type'] = "File"
    elif handler_type == "Int":
        inp['type'] = "Number"
        inp['integer'] = True
    elif handler_type == "Float":
        inp['type'] = "Number"
    elif handler_type == "Bool":
        inp['type'] = "Flag"
    else:
        inp['type'] = "String"

    # Deal with range inputs
    if handler_type == "Range":
        inp['type'] = "Number"
        if trait_handler._low is not None:
            inp['minimum'] = trait_handler._low
        if trait_handler._high is not None:
            inp['maximum'] = trait_handler._high
        if trait_handler._exclude_low:
            inp['exclusive-minimum'] = True
        if trait_handler._exclude_high:
            inp['exclusive-maximum'] = True

    # Deal with list inputs
    # TODO handle lists of lists (e.g. FSL ProbTrackX seed input)
    if handler_type == "List":
        inp['list'] = True
        item_type = trait_handler.item_trait.trait_type
        item_type_name = type(item_type).__name__
        if item_type_name == "Int":
            inp['integer'] = True
            inp['type'] = "Number"
        elif item_type_name == "Float":
            inp['type'] = "Number"
        elif item_type_name == "File":
            inp['type'] = "File"
        elif item_type_name == "Enum":
            value_choices = item_type.values
            if value_choices is not None:
                if all(isinstance(n, int) for n in value_choices):
                    inp['type'] = "Number"
                    inp['integer'] = True
                elif all(isinstance(n, float) for n in value_choices):
                    inp['type'] = "Number"
                inp['value-choices'] = value_choices
        else:
            inp['type'] = "String"
        if trait_handler.minlen != 0:
            inp['min-list-entries'] = trait_handler.minlen
        if trait_handler.maxlen != sys.maxsize:
            inp['max-list-entries'] = trait_handler.maxlen
        if spec.sep:
            inp['list-separator'] = spec.sep

    if handler_type == "Tuple":
        inp['list'] = True
        inp['min-list-entries'] = len(spec.default)
        inp['max-list-entries'] = len(spec.default)
        input_type = type(spec.default[0]).__name__
        if input_type == 'int':
            inp['type'] = "Number"
            inp['integer'] = True
        elif input_type == 'float':
            inp['type'] = "Number"
        else:
            inp['type'] = "String"

    # Deal with multi-input
    if handler_type == "InputMultiObject":
        inp['type'] = "File"
        inp['list'] = True
        if spec.sep:
            inp['list-separator'] = spec.sep

    inp['value-key'] = "[" + input_name.upper(
    ) + "]"  # assumes that input names are unique

    flag, flag_sep = get_command_line_flag(spec, inp['type'] == "Flag",
                                           input_name)

    if flag is not None:
        inp['command-line-flag'] = flag
    if flag_sep is not None:
        inp['command-line-flag-separator'] = flag_sep

    inp['description'] = get_description_from_spec(inputs, input_name, spec)
    if not (hasattr(spec, "mandatory") and spec.mandatory):
        inp['optional'] = True
    else:
        inp['optional'] = False
    if spec.usedefault:
        inp['default-value'] = spec.default_value()[1]
    if spec.requires is not None:
        inp['requires-inputs'] = spec.requires

    try:
        value_choices = trait_handler.values
    except AttributeError:
        pass
    else:
        if value_choices is not None:
            if all(isinstance(n, int) for n in value_choices):
                inp['type'] = "Number"
                inp['integer'] = True
            elif all(isinstance(n, float) for n in value_choices):
                inp['type'] = "Number"
            inp['value-choices'] = value_choices

    return inp


def get_boutiques_output(outputs, name, spec, interface, tool_inputs):
    """
    Returns a dictionary containing the Boutiques output corresponding
    to a Nipype output.

    Args:
      * outputs: outputs of the Nipype interface.
      * name: name of the Nipype output.
      * spec: Nipype output spec.
      * interface: Nipype interface.
      * tool_inputs: list of tool inputs (as produced by method
        get_boutiques_input).

    Assumes that:
      * Output names are unique.
      * Input values involved in the path template are defined.
      * Output files are written in the current directory.
      * There is a single output value (output lists are not supported).
    """
    output = {}
    output['name'] = name.replace('_', ' ').capitalize()

    # Check if the output name was already used as an input name
    # If so, append '_outfile' to the end of the ID
    unique_id = True
    for inp in tool_inputs:
        if inp['id'] == name:
            unique_id = False
            break
    output['id'] = name if unique_id else name + '_outfile'

    output['path-template'] = ""

    # No real way to determine if an output is always
    # produced, regardless of the input values.
    output['optional'] = True

    output['description'] = get_description_from_spec(outputs, name, spec)

    # Path template creation.

    try:
        output_value = interface._list_outputs()[name]
    except TypeError:
        output_value = None
    except AttributeError:
        output_value = None
    except KeyError:
        output_value = None

    # Handle multi-outputs
    if (isinstance(output_value, list) or
            type(spec.handler).__name__ == "OutputMultiObject" or
            type(spec.handler).__name__ == "List"):
        output['list'] = True
        if output_value:
            # Check if all extensions are the same
            extensions = []
            for val in output_value:
                extensions.append(os.path.splitext(val)[1])
            # If extensions all the same, set path template as
            # wildcard + extension. Otherwise just use a wildcard
            if len(set(extensions)) == 1:
                output['path-template'] = "*" + extensions[0]
            else:
                output['path-template'] = "*"
            return output

    # If an output value is defined, use its relative path, if one exists.
    # Otherwise, put blank string as placeholder and try to fill it on
    # another iteration.
    if output_value:
        output['path-template'] = os.path.relpath(output_value)
    else:
        output['path-template'] = ""

    return output


def get_boutiques_groups(input_traits):
    """
    Returns a list of dictionaries containing Boutiques groups for the mutually
    exclusive Nipype inputs.
    """
    desc_groups = []
    mutex_input_sets = []

    # Get all the groups
    for name, spec in input_traits:
        if spec.xor is not None:
            group_members = set([name] + list(spec.xor))
            if group_members not in mutex_input_sets:
                mutex_input_sets.append(group_members)

    # Create a dictionary for each one
    for i, inp_set in enumerate(mutex_input_sets, 1):
        desc_groups.append({'id': "mutex_group" +
                                  ("_" + str(i) if i != 1 else ""),
                            'name': "Mutex group" +
                                    (" " + str(i) if i != 1 else ""),
                            'members': list(inp_set),
                            'mutually-exclusive': True})

    return desc_groups


def get_description_from_spec(obj, name, spec):
    '''
    Generates a description based on the input or output spec.
    '''
    if not spec.desc:
        spec.desc = "No description provided."
    spec_info = spec.full_info(obj, name, None)

    boutiques_description = (spec_info.capitalize(
    ) + ". " + spec.desc.capitalize()).replace("\n", '')

    if not boutiques_description.endswith('.'):
        boutiques_description += '.'

    return boutiques_description


def fill_in_missing_output_path(output, output_name, tool_inputs):
    '''
    Creates a path template for outputs that are missing one
    This is needed for the descriptor to be valid (path template is required)
    '''
    # Look for an input with the same name as the output and use its value key
    found = False
    for input in tool_inputs:
        if input['name'] == output_name:
            output['path-template'] = input['value-key']
            found = True
            break
    # If no input with the same name was found, use the output ID
    if not found:
        output['path-template'] = output['id']
    return output


def generate_custom_inputs(desc_inputs):
    '''
    Generates a bunch of custom input dictionaries in order to generate
    as many outputs as possible (to get their path templates).
    Currently only works with flag inputs and inputs with defined value
    choices.
    '''
    custom_input_dicts = []
    for desc_input in desc_inputs:
        if desc_input['type'] == 'Flag':
            custom_input_dicts.append({desc_input['id']: True})
        elif desc_input.get('value-choices') and not desc_input.get('list'):
            for value in desc_input['value-choices']:
                custom_input_dicts.append({desc_input['id']: value})
    return custom_input_dicts


def reorder_cmd_line_args(cmd_line, interface, ignore_inputs=None):
    '''
    Generates a new command line with the positional arguments in the
    correct order
    '''
    interface_name = cmd_line.split()[0]
    positional_arg_dict = {}
    positional_args = []
    non_positional_args = []

    for name, spec in sorted(interface.inputs.traits(transient=None).items()):
        if ignore_inputs is not None and name in ignore_inputs:
            continue
        value_key = "[" + name.upper() + "]"
        if spec.position is not None:
            positional_arg_dict[spec.position] = value_key
        else:
            non_positional_args.append(value_key)

    last_arg = None
    for item in sorted(positional_arg_dict.items()):
        if item[0] == -1:
            last_arg = item[1]
            continue
        positional_args.append(item[1])

    return (interface_name + " " +
            ((" ".join(positional_args) + " ")
                if len(positional_args) > 0 else "") +
            ((last_arg + " ") if last_arg else "") +
            " ".join(non_positional_args))


def get_command_line_flag(input_spec, is_flag_type=False, input_name=None):
    '''
    Generates the command line flag for a given input
    '''
    flag, flag_sep = None, None
    if input_spec.argstr:
        if "=" in input_spec.argstr:
            if (input_spec.argstr.split("=")[1] == '0'
                    or input_spec.argstr.split("=")[1] == '1'):
                flag = input_spec.argstr
            else:
                flag = input_spec.argstr.split("=")[0].strip()
                flag_sep = "="
        elif input_spec.argstr.split("%")[0]:
            flag = input_spec.argstr.split("%")[0].strip()
    elif is_flag_type:
        flag = ("--%s" % input_name + " ").strip()
    return flag, flag_sep


def get_boutiques_output_from_inp(inputs, inp_spec, inp_name):
    '''
    Takes a Nipype input representing an output file and generates a
    Boutiques output for it
    '''
    output = {}
    output['name'] = inp_name.replace('_', ' ').capitalize()
    output['id'] = inp_name
    output['optional'] = True
    output['description'] = get_description_from_spec(inputs, inp_name,
                                                      inp_spec)
    if not (hasattr(inp_spec, "mandatory") and inp_spec.mandatory):
        output['optional'] = True
    else:
        output['optional'] = False
    if inp_spec.usedefault:
        output['default-value'] = inp_spec.default_value()[1]
    if isinstance(inp_spec.name_source, list):
        source = inp_spec.name_source[0]
    else:
        source = inp_spec.name_source
    output['path-template'] = inp_spec.name_template.replace(
        "%s", "[" + source.upper() + "]")
    output['value-key'] = "[" + inp_name.upper() + "]"
    flag, flag_sep = get_command_line_flag(inp_spec)
    if flag is not None:
        output['command-line-flag'] = flag
    if flag_sep is not None:
        output['command-line-flag-separator'] = flag_sep
    return output



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utilities import GeneratePurePlugMask


def test_GeneratePurePlugMask_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputImageModalities=dict(argstr='--inputImageModalities %s...', ),
        numberOfSubSamples=dict(
            argstr='--numberOfSubSamples %s',
            sep=',',
        ),
        outputMaskFile=dict(
            argstr='--outputMaskFile %s',
            hash_files=False,
        ),
        threshold=dict(argstr='--threshold %f', ),
    )
    inputs = GeneratePurePlugMask.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_GeneratePurePlugMask_outputs():
    output_map = dict(outputMaskFile=dict(extensions=None, ), )
    outputs = GeneratePurePlugMask.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..minc import ToRaw


def test_ToRaw_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        input_file=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        nonormalize=dict(
            argstr='-nonormalize',
            xor=('normalize', 'nonormalize'),
        ),
        normalize=dict(
            argstr='-normalize',
            xor=('normalize', 'nonormalize'),
        ),
        out_file=dict(
            argstr='> %s',
            extensions=None,
            genfile=True,
            position=-1,
        ),
        output_file=dict(
            extensions=None,
            hash_files=False,
            keep_extension=False,
            name_source=['input_file'],
            name_template='%s.raw',
            position=-1,
        ),
        write_byte=dict(
            argstr='-byte',
            xor=('write_byte', 'write_short', 'write_int', 'write_long',
                 'write_float', 'write_double'),
        ),
        write_double=dict(
            argstr='-double',
            xor=('write_byte', 'write_short', 'write_int', 'write_long',
                 'write_float', 'write_double'),
        ),
        write_float=dict(
            argstr='-float',
            xor=('write_byte', 'write_short', 'write_int', 'write_long',
                 'write_float', 'write_double'),
        ),
        write_int=dict(
            argstr='-int',
            xor=('write_byte', 'write_short', 'write_int', 'write_long',
                 'write_float', 'write_double'),
        ),
        write_long=dict(
            argstr='-long',
            xor=('write_byte', 'write_short', 'write_int', 'write_long',
                 'write_float', 'write_double'),
        ),
        write_range=dict(argstr='-range %s %s', ),
        write_short=dict(
            argstr='-short',
            xor=('write_byte', 'write_short', 'write_int', 'write_long',
                 'write_float', 'write_double'),
        ),
        write_signed=dict(
            argstr='-signed',
            xor=('write_signed', 'write_unsigned'),
        ),
        write_unsigned=dict(
            argstr='-unsigned',
            xor=('write_signed', 'write_unsigned'),
        ),
    )
    inputs = ToRaw.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_ToRaw_outputs():
    output_map = dict(output_file=dict(extensions=None, ), )
    outputs = ToRaw.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import Overlay


def test_Overlay_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        auto_thresh_bg=dict(
            argstr='-a',
            mandatory=True,
            position=5,
            xor=('auto_thresh_bg', 'full_bg_range', 'bg_thresh'),
        ),
        background_image=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=4,
        ),
        bg_thresh=dict(
            argstr='%.3f %.3f',
            mandatory=True,
            position=5,
            xor=('auto_thresh_bg', 'full_bg_range', 'bg_thresh'),
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        full_bg_range=dict(
            argstr='-A',
            mandatory=True,
            position=5,
            xor=('auto_thresh_bg', 'full_bg_range', 'bg_thresh'),
        ),
        out_file=dict(
            argstr='%s',
            extensions=None,
            genfile=True,
            hash_files=False,
            position=-1,
        ),
        out_type=dict(
            argstr='%s',
            position=2,
            usedefault=True,
        ),
        output_type=dict(),
        show_negative_stats=dict(
            argstr='%s',
            position=8,
            xor=['stat_image2'],
        ),
        stat_image=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=6,
        ),
        stat_image2=dict(
            argstr='%s',
            extensions=None,
            position=9,
            xor=['show_negative_stats'],
        ),
        stat_thresh=dict(
            argstr='%.2f %.2f',
            mandatory=True,
            position=7,
        ),
        stat_thresh2=dict(
            argstr='%.2f %.2f',
            position=10,
        ),
        transparency=dict(
            argstr='%s',
            position=1,
            usedefault=True,
        ),
        use_checkerboard=dict(
            argstr='-c',
            position=3,
        ),
    )
    inputs = Overlay.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Overlay_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = Overlay.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..epi import TOPUP


def test_TOPUP_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        config=dict(
            argstr='--config=%s',
            usedefault=True,
        ),
        encoding_direction=dict(
            argstr='--datain=%s',
            mandatory=True,
            requires=['readout_times'],
            xor=['encoding_file'],
        ),
        encoding_file=dict(
            argstr='--datain=%s',
            extensions=None,
            mandatory=True,
            xor=['encoding_direction'],
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        estmov=dict(argstr='--estmov=%d', ),
        fwhm=dict(argstr='--fwhm=%f', ),
        in_file=dict(
            argstr='--imain=%s',
            extensions=None,
            mandatory=True,
        ),
        interp=dict(argstr='--interp=%s', ),
        max_iter=dict(argstr='--miter=%d', ),
        minmet=dict(argstr='--minmet=%d', ),
        numprec=dict(argstr='--numprec=%s', ),
        out_base=dict(
            argstr='--out=%s',
            extensions=None,
            hash_files=False,
            name_source=['in_file'],
            name_template='%s_base',
        ),
        out_corrected=dict(
            argstr='--iout=%s',
            extensions=None,
            hash_files=False,
            name_source=['in_file'],
            name_template='%s_corrected',
        ),
        out_field=dict(
            argstr='--fout=%s',
            extensions=None,
            hash_files=False,
            name_source=['in_file'],
            name_template='%s_field',
        ),
        out_jac_prefix=dict(
            argstr='--jacout=%s',
            hash_files=False,
            usedefault=True,
        ),
        out_logfile=dict(
            argstr='--logout=%s',
            extensions=None,
            hash_files=False,
            keep_extension=True,
            name_source=['in_file'],
            name_template='%s_topup.log',
        ),
        out_mat_prefix=dict(
            argstr='--rbmout=%s',
            hash_files=False,
            usedefault=True,
        ),
        out_warp_prefix=dict(
            argstr='--dfout=%s',
            hash_files=False,
            usedefault=True,
        ),
        output_type=dict(),
        readout_times=dict(
            mandatory=True,
            requires=['encoding_direction'],
            xor=['encoding_file'],
        ),
        reg_lambda=dict(argstr='--lambda=%0.f', ),
        regmod=dict(argstr='--regmod=%s', ),
        regrid=dict(argstr='--regrid=%d', ),
        scale=dict(argstr='--scale=%d', ),
        splineorder=dict(argstr='--splineorder=%d', ),
        ssqlambda=dict(argstr='--ssqlambda=%d', ),
        subsamp=dict(argstr='--subsamp=%d', ),
        warp_res=dict(argstr='--warpres=%f', ),
    )
    inputs = TOPUP.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_TOPUP_outputs():
    output_map = dict(
        out_corrected=dict(extensions=None, ),
        out_enc_file=dict(extensions=None, ),
        out_field=dict(extensions=None, ),
        out_fieldcoef=dict(extensions=None, ),
        out_jacs=dict(),
        out_logfile=dict(extensions=None, ),
        out_mats=dict(),
        out_movpar=dict(extensions=None, ),
        out_warps=dict(),
    )
    outputs = TOPUP.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value
